import numpy as np
import pytest
from astropy import convolution
from scipy.signal import medfilt
import astropy.units as u
from astropy.nddata import StdDevUncertainty, VarianceUncertainty, InverseVariance

from ..spectra import Spectrum
from ..manipulation.smoothing import (convolution_smooth, box_smooth,
                                      gaussian_smooth, trapezoid_smooth,
                                      median_smooth)


def compare_flux(flux_smooth1, flux_smooth2, flux_original, rtol=0.01):
    """
    There are two things to compare for each set of smoothing:

    1. Compare the smoothed flux from the astropy machinery vs
       the smoothed flux from specutils.  This is done by
       comparing flux_smooth1 and flux_smooth2.

    2. Next we want to compare the smoothed flux to the original
       flux.  This is a little more difficult as smoothing will
       make a difference for median filter, but less so for
       convolution based smoothing if the kernel is normalized
       (area under the kernel = 1).

       In this second case the rtol (relative tolerance) is used
       judiciously.

    """

    # Compare, element by element, the two smoothed fluxes.
    # Workaround for astropy dev (4.2) which seems to return quantities in
    #  convolutions
    if isinstance(flux_smooth1, u.Quantity):
        flux_smooth1 = flux_smooth1.value
    if isinstance(flux_smooth2, u.Quantity):
        flux_smooth2 = flux_smooth2.value

    assert np.allclose(flux_smooth1, flux_smooth2)

    # Compare the total spectral flux of the smoothed to the original.
    assert np.allclose(sum(flux_smooth1), sum(flux_original), rtol=rtol)


def test_smooth_custom_kernel(simulated_spectra):
    """
    Test CustomKernel smoothing with correct parameters.
    """

    # Create the original spectrum
    spec1 = simulated_spectra.s1_um_mJy_e1
    flux_original = spec1.flux

    # Create a custom kernel (some weird asymmetric-ness)
    numpy_kernel = np.array([0.5, 1, 2, 0.5, 0.2])
    numpy_kernel = numpy_kernel / np.sum(numpy_kernel)

    custom_kernel = convolution.CustomKernel(numpy_kernel)
    flux_smoothed_astropy = convolution.convolve(flux_original, custom_kernel)

    # Calculate the custom smoothed
    spec1_smoothed = convolution_smooth(spec1, custom_kernel)
    compare_flux(spec1_smoothed.flux.value, flux_smoothed_astropy, flux_original.value)


@pytest.mark.parametrize("width", [1, 2.3])
def test_smooth_box_good(simulated_spectra, width):
    """
    Test Box1DKernel smoothing with correct parameters.

    Width values need to be a number greater than 0.
    """

    # Create the original spectrum
    spec1 = simulated_spectra.s1_um_mJy_e1
    flux_original = spec1.flux

    # Calculate the smoothed flux using Astropy
    box_kernel = convolution.Box1DKernel(width)
    flux_smoothed_astropy = convolution.convolve(flux_original, box_kernel)

    # Calculate the box smoothed
    spec1_smoothed = box_smooth(spec1, width)
    compare_flux(spec1_smoothed.flux.value, flux_smoothed_astropy, flux_original.value)

    # Check the input and output units
    assert spec1.wavelength.unit == spec1_smoothed.wavelength.unit
    assert spec1.flux.unit == spec1_smoothed.flux.unit
    assert len(spec1.meta) == len(spec1_smoothed.meta)


@pytest.mark.parametrize("width", [-1, 0, 'a'])
def test_smooth_box_bad(simulated_spectra, width):
    """
    Test Box1DKernel smoothing with incorrect parameters.

    Width values need to be a number greater than 0.
    """

    #  Create the spectrum
    spec1 = simulated_spectra.s1_um_mJy_e1

    # Test bad input parameters
    with pytest.raises(ValueError):
        box_smooth(spec1, width)


@pytest.mark.parametrize("stddev", [1, 2.3])
def test_smooth_gaussian_good(simulated_spectra, stddev):
    """
    Test Gaussian1DKernel smoothing with correct parameters.

    Standard deviation values need to be a number greater than 0.
    """

    #  Create the spectrum
    spec1 = simulated_spectra.s1_um_mJy_e1
    flux_original = spec1.flux

    # Calculate the smoothed flux using Astropy
    gaussian_kernel = convolution.Gaussian1DKernel(stddev)
    flux_smoothed_astropy = convolution.convolve(flux_original, gaussian_kernel)

    # Test gaussian smoothing
    spec1_smoothed = gaussian_smooth(spec1, stddev)
    compare_flux(spec1_smoothed.flux.value, flux_smoothed_astropy, flux_original.value, rtol=0.02)

    # Check the input and output units
    assert spec1.wavelength.unit == spec1_smoothed.wavelength.unit
    assert spec1.flux.unit == spec1_smoothed.flux.unit
    assert len(spec1.meta) == len(spec1_smoothed.meta)


@pytest.mark.parametrize("stddev", [-1, 0, 'a'])
def test_smooth_gaussian_bad(simulated_spectra, stddev):
    """
    Test MexicanHat1DKernel smoothing with incorrect parameters.

    Standard deviation values need to be a number greater than 0.
    """

    #  Create the spectrum
    spec1 = simulated_spectra.s1_um_mJy_e1

    # Test bad input paramters
    with pytest.raises(ValueError):
        gaussian_smooth(spec1, stddev)


@pytest.mark.parametrize("stddev", [1, 2.3])
def test_smooth_trapezoid_good(simulated_spectra, stddev):
    """
    Test Trapezoid1DKernel smoothing with correct parameters.

    Standard deviation values need to be a number greater than 0.
    """

    #  Create the spectrum
    spec1 = simulated_spectra.s1_um_mJy_e1
    flux_original = spec1.flux

    # Create the flux_smoothed which is what we want to compare to
    trapezoid_kernel = convolution.Trapezoid1DKernel(stddev)
    flux_smoothed_astropy = convolution.convolve(flux_original, trapezoid_kernel)

    # Test trapezoid smoothing
    spec1_smoothed = trapezoid_smooth(spec1, stddev)
    compare_flux(spec1_smoothed.flux.value, flux_smoothed_astropy, flux_original.value)

    # Check the input and output units
    assert spec1.wavelength.unit == spec1_smoothed.wavelength.unit
    assert spec1.flux.unit == spec1_smoothed.flux.unit
    assert len(spec1.meta) == len(spec1_smoothed.meta)


@pytest.mark.parametrize("stddev", [-1, 0, 'a'])
def test_smooth_trapezoid_bad(simulated_spectra, stddev):
    """
    Test Trapezoid1DKernel smoothing with incorrect parameters.

    Standard deviation values need to be a number greater than 0.
    """

    #  Create the spectrum
    spec1 = simulated_spectra.s1_um_mJy_e1

    # Test bad parameters
    with pytest.raises(ValueError):
        trapezoid_smooth(spec1, stddev)


@pytest.mark.parametrize("width", [1, 3, 9])
def test_smooth_median_good(simulated_spectra, width):
    """
    Test Median smoothing with correct parameters.

    Width values need to be a number greater than 0.
    """

    #  Create the spectrum
    spec1 = simulated_spectra.s1_um_mJy_e1
    flux_original = spec1.flux

    # Create the flux_smoothed which is what we want to compare to
    flux_smoothed_astropy = medfilt(flux_original, width)

    # Test median smoothing
    spec1_smoothed = median_smooth(spec1, width)
    compare_flux(spec1_smoothed.flux.value, flux_smoothed_astropy, flux_original.value, rtol=0.15)

    # Check the input and output units
    assert spec1.wavelength.unit == spec1_smoothed.wavelength.unit
    assert spec1.flux.unit == spec1_smoothed.flux.unit
    assert len(spec1.meta) == len(spec1_smoothed.meta)


@pytest.mark.parametrize("width", [-1, 0, 'a'])
def test_smooth_median_bad(simulated_spectra, width):
    """
    Test Median smoothing with incorrect parameters.

    Width values need to be a number greater than 0.
    """

    #  Create the spectrum
    spec1 = simulated_spectra.s1_um_mJy_e1

    # Test bad parameters
    with pytest.raises(ValueError):
        median_smooth(spec1, width)


def test_smooth_custom_kernel_uncertainty(simulated_spectra):
    """
    Test CustomKernel smoothing with correct parameters.
    """

    np.random.seed(42)

    # Create a custom kernel (some weird asymmetric-ness)
    numpy_kernel = np.array([0.5, 1, 2, 0.5, 0.2])
    numpy_kernel = numpy_kernel / np.sum(numpy_kernel)
    custom_kernel = convolution.CustomKernel(numpy_kernel)

    spec1 = simulated_spectra.s1_um_mJy_e1
    uncertainty = np.abs(np.random.random(spec1.flux.shape))

    # Test StdDevUncertainty
    spec1.uncertainty = StdDevUncertainty(uncertainty)

    spec1_smoothed = convolution_smooth(spec1, custom_kernel)
    tt = convolution.convolve(1/(spec1.uncertainty.array**2), custom_kernel)
    uncertainty_smoothed_astropy = 1/np.sqrt(tt)

    assert np.allclose(spec1_smoothed.uncertainty.array, uncertainty_smoothed_astropy)

    # Test VarianceUncertainty
    spec1.uncertainty = VarianceUncertainty(uncertainty)

    spec1_smoothed = convolution_smooth(spec1, custom_kernel)
    uncertainty_smoothed_astropy = 1/convolution.convolve(1/spec1.uncertainty.array, custom_kernel)
    assert np.allclose(spec1_smoothed.uncertainty.array, uncertainty_smoothed_astropy)

    # Test InverseVariance
    spec1.uncertainty = InverseVariance(uncertainty)

    spec1_smoothed = convolution_smooth(spec1, custom_kernel)
    uncertainty_smoothed_astropy = convolution.convolve(spec1.uncertainty.array, custom_kernel)
    assert np.allclose(spec1_smoothed.uncertainty.array, uncertainty_smoothed_astropy)


def test_medfilt_dtype():
    """
    Check that dtypes not supported by scipy's medfilt method are
    converted to supported dtypes before they are passed to scipy.
    """
    flux = [1, 2, 3] * u.Jy

    for dtype in ('>f4', np.float32, 'i4', int):
        spectrum = Spectrum(flux=flux.astype(dtype), spectral_axis=[1, 2, 3] * u.um)
        median_smooth(spectrum, 1)



import pytest
import os
import urllib


remote_access = lambda argvals: pytest.mark.parametrize(
    'remote_data_path', argvals, indirect=True, scope='function')


@pytest.fixture(scope='module')
def remote_data_path(request, tmp_path):
    """
    Remotely access the Zenodo deposition archive to retrieve the versioned
    test data.
    """
    # Make use of configuration option from pytest-remotedata in order to
    # control access to remote data.
    if request.config.getoption('remote_data', 'any') != 'any':
        pytest.skip()

    file_id, file_name = request.param.values()

    url = "https://zenodo.org/record/{}/files/{}?download=1".format(
        file_id, file_name)

    # Create a temporary directory that is automatically cleaned up when the
    # context is exited, removing any temporarily stored data.
    file_path = os.path.join(tmp_path, file_name)

    with urllib.request.urlopen(url) as r, open(file_path, 'wb') as tmp_file:
        tmp_file.write(r.read())

        yield file_path



import astropy.units as u
import numpy as np
from numpy.testing import assert_allclose
import pytest
from astropy.nddata import StdDevUncertainty
from astropy.coordinates import SpectralCoord
from gwcs.wcs import WCS as GWCS

from ..spectra.spectrum import Spectrum
from ..spectra.spectrum_collection import SpectrumCollection
from ..utils.wcs_utils import gwcs_from_array


@pytest.fixture
def spectrum_collection():
    flux = u.Quantity(np.random.sample((5, 10)), unit='Jy')
    spectral_axis = u.Quantity(np.arange(50).reshape((5, 10)) + 1, unit='AA')
    wcs = np.array([gwcs_from_array(x, flux.shape, spectral_axis_index=1) for x in spectral_axis])
    uncertainty = StdDevUncertainty(np.random.sample((5, 10)), unit='Jy')
    mask = np.ones((5, 10)).astype(bool)
    meta = [{'test': 5, 'info': [1, 2, 3]} for i in range(5)]

    spec_coll = SpectrumCollection(
        flux=flux, spectral_axis=spectral_axis, wcs=wcs,
        uncertainty=uncertainty, mask=mask, meta=meta,
        spectral_axis_index=1)

    return spec_coll


def test_create_spectrum_collection(spectrum_collection):
    assert spectrum_collection.ndim == 1
    assert spectrum_collection.shape == (5, )
    assert spectrum_collection.nspectral == 10
    assert isinstance(spectrum_collection.flux, u.Quantity)
    assert isinstance(spectrum_collection.spectral_axis, u.Quantity)
    assert spectrum_collection.wavelength.unit.physical_type == 'length'
    assert spectrum_collection.frequency.unit.physical_type == 'frequency'
    assert spectrum_collection.energy.unit.physical_type == 'energy'
    assert isinstance(spectrum_collection.uncertainty, StdDevUncertainty)
    assert spectrum_collection.uncertainty.array.shape == spectrum_collection.flux.shape


def test_spectrum_collection_slicing(spectrum_collection):
    # Test that slicing a spectrum collection returns a Spectrum
    single_spec = spectrum_collection[0]

    assert isinstance(single_spec, Spectrum)
    assert isinstance(single_spec.flux, u.Quantity)
    assert isinstance(single_spec.spectral_axis, u.Quantity)
    assert single_spec.flux.shape == spectrum_collection.flux.shape[1:]
    assert isinstance(single_spec.wcs, GWCS)
    assert single_spec.mask.shape == single_spec.flux.shape
    assert len(single_spec.meta) == len(spectrum_collection.meta[0])
    assert isinstance(single_spec.uncertainty, StdDevUncertainty)


def test_collection_without_optional_arguments():
    # Without uncertainties
    flux = u.Quantity(np.random.sample((5, 10)), unit='Jy')
    spectral_axis = u.Quantity(np.arange(50).reshape((5, 10)) + 1, unit='AA')
    uncertainty = StdDevUncertainty(np.random.sample((5, 10)), unit='Jy')
    wcs = np.array([gwcs_from_array(x, flux.shape, spectral_axis_index=1) for x in spectral_axis])
    mask = np.ones((5, 10)).astype(bool)
    meta = [{'test': 5, 'info': [1, 2, 3]} for i in range(5)]

    spec_coll = SpectrumCollection(
        flux=flux, spectral_axis=spectral_axis, wcs=wcs, mask=mask, meta=meta)

    # Without mask
    spec_coll = SpectrumCollection(
        flux=flux, spectral_axis=spectral_axis, uncertainty=uncertainty,
        wcs=wcs, meta=meta)

    # Without meta
    spec_coll = SpectrumCollection(  # noqa
        flux=flux, spectral_axis=spectral_axis, uncertainty=uncertainty,
        wcs=wcs, mask=mask)


@pytest.mark.filterwarnings('ignore:Not all spectra have associated masks')
def test_create_collection_from_spectrum():
    spec = Spectrum(spectral_axis=SpectralCoord(np.linspace(0, 50, 50) * u.AA,
                      redshift=0.1), flux=np.random.randn(50) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))
    spec1 = Spectrum(spectral_axis=SpectralCoord(np.linspace(20, 60, 50) * u.AA,
                       redshift=0.1), flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    spec_coll = SpectrumCollection.from_spectra([spec, spec1])

    assert spec_coll.ndim == 1
    assert spec_coll.shape == (2, )
    assert spec_coll.nspectral == 50
    assert isinstance(spec_coll.flux, u.Quantity)
    assert isinstance(spec_coll.spectral_axis, SpectralCoord)
    assert spec.spectral_axis.unit == spec_coll.spectral_axis.unit
    assert spec.flux.unit == spec_coll.flux.unit
    assert_allclose(spec_coll.spectral_axis.redshift.value, 0.1)


@pytest.mark.filterwarnings('ignore:Not all spectra have associated masks')
def test_create_collection_from_collections():
    spec = Spectrum(spectral_axis=np.linspace(0, 50, 50) * u.AA,
                      flux=np.random.randn(50) * u.Jy,
                      uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))
    spec1 = Spectrum(spectral_axis=np.linspace(20, 60, 50) * u.AA,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    spec_coll1 = SpectrumCollection.from_spectra([spec, spec1])

    spec2 = Spectrum(spectral_axis=np.linspace(40, 80, 50) * u.AA,
                       flux=np.random.randn(50) * u.Jy,
                       uncertainty=StdDevUncertainty(np.random.sample(50), unit='Jy'))

    spec_coll2 = SpectrumCollection.from_spectra([spec, spec2])

    spec_coll = SpectrumCollection.from_spectra([spec_coll1, spec_coll2, spec_coll1])

    assert spec_coll.ndim == 2
    assert spec_coll.shape == (3, 2)
    assert spec_coll.nspectral == 50
    assert isinstance(spec_coll.flux, u.Quantity)
    assert isinstance(spec_coll.spectral_axis, u.Quantity)
    assert spec.spectral_axis.unit == spec_coll.spectral_axis.unit
    assert spec.flux.unit == spec_coll.flux.unit


@pytest.mark.filterwarnings('ignore:Not all spectra have associated uncertainties of the same type')
@pytest.mark.filterwarnings('ignore:Not all spectra have associated masks')
def test_create_collection_from_spectra_without_uncertainties():
    spec = Spectrum(spectral_axis=np.linspace(0, 50, 50) * u.AA,
                      flux=np.random.randn(50) * u.Jy)
    spec1 = Spectrum(spectral_axis=np.linspace(20, 60, 50) * u.AA,
                       flux=np.random.randn(50) * u.Jy)

    SpectrumCollection.from_spectra([spec, spec1])


def test_mismatched_spectral_axes_parameters():
    spec = Spectrum(spectral_axis=SpectralCoord(np.linspace(0, 50, 50) * u.AA,
                      radial_velocity=u.Quantity(100.0, "km/s")),
                      flux=np.random.randn(50) * u.Jy)
    spec1 = Spectrum(spectral_axis=SpectralCoord(np.linspace(20, 60, 50) * u.AA,
                       radial_velocity=u.Quantity(200.0, "km/s")),
                       flux=np.random.randn(50) * u.Jy)

    with pytest.raises(ValueError):
        SpectrumCollection.from_spectra([spec, spec1])


@pytest.mark.parametrize('scshape,expected_len', [((5, 10), 5), ((4, 5, 10), 4)])
def test_len(scshape, expected_len):
    flux = u.Quantity(np.random.sample(scshape), unit='Jy')
    spectral_axis = u.Quantity(np.arange(np.prod(scshape)).reshape(scshape) + 1, unit='AA')
    sc2d = SpectrumCollection(flux=flux, spectral_axis=spectral_axis)

    assert sc2d.shape == scshape[:-1]
    assert len(sc2d) == expected_len



from astropy.nddata import VarianceUncertainty
from astropy.table import QTable
from specutils import SpectrumList
from specutils.io.default_loaders.dc_common import (
    FITS_FILE_EXTS, add_single_spectra_to_map,
)
from specutils.io.parsing_utils import read_fileobj_or_hdulist
from specutils.io.registers import data_loader

# There appears to be nothing which says "this is a SAMI 1D spectra", so guess
# it based on the headers that should be there
SAMI_1D_SPECTRA_HEADER_KEYWORDS = [
    "BUNIT", "CATADEC", "CATARA", "CDELT1", "CRPIX1", "CRVAL1", "CTYPE1",
    "CUNIT1", "DROPFACT", "ELLIP", "GRATID", "IFUPROBE", "KPC_SIZE", "NAME",
    "N_SPAX", "POS_ANG", "PSFALPHA", "PSFBETA", "PSFFWHM", "RADESYS", "RADIUS",
    "RO_GAIN", "RO_NOISE", "STDNAME", "WCSAXES", "Z_TONRY",
]


def identify_sami_cube(origin, *args, **kwargs):
    """
    Identify if the current file is a SAMI cube file
    """
    # TODO check this
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        header = hdulist[0].header
        data = hdulist[0].data
        if "SAMI" in header.get("INSTRUME", "") and len(data.shape) == 3:
            return True
        return False


def identify_sami_1d_spec(origin, *args, **kwargs):
    """
    Identify if the current file is a SAMI 1d spectra file of some kind
    """
    # TODO check this
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        header = hdulist[0].header
        for key in SAMI_1D_SPECTRA_HEADER_KEYWORDS:
            if key not in header:
                return False
        return True


@data_loader(
    label="SAMI-cube", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_sami_cube, priority=10,
)
def sami_cube_loader(filename):
    spectra_map = {
        "sky": [],
        "combined": [],
        "unreduced": [],
        "reduced": [],
    }
    primary_header = None

    with read_fileobj_or_hdulist(filename) as hdulist:
        for i, hdu in enumerate(hdulist):
            if i == 0:
                # This is the primary extension, and the one with the
                # science data. The header is fairly complete.
                primary_header = hdu.header
                spec = add_single_spectra_to_map(
                    spectra_map,
                    header=primary_header,
                    data=hdu.data,
                    index=None,
                    all_standard_units=True,
                    all_keywords=True,
                    valid_wcs=True,
                )

            elif "VARIANCE" == hdu.header.get("EXTNAME"):
                # This is the variance extension, and is missing wcs and
                # units.
                uncertainty = VarianceUncertainty(
                    hdu.data, unit=spec.flux.unit ** 2
                )
                spec.uncertainty = uncertainty

            elif "WEIGHT" == hdu.header.get("EXTNAME"):
                # This is the weight extension, and is missing wcs. The
                # units are effectively "normalised" (from 0-1 it seems).
                spec.meta["sami_cube_weight_map"] = hdu.data

            elif "COVAR" == hdu.header.get("EXTNAME"):
                # This is the spatial covariance extension. It's not clear
                # as to how best to expose this, so skipping for now.
                pass

            elif "QC" == hdu.header.get("EXTNAME"):
                # This is the QC extension, and is a binary table. This we
                # add to the metadata.
                spec.meta["sami_QC_table"] = QTable.read(hdu)

            elif "DUST" == hdu.header.get("EXTNAME"):
                # This is the dust extension, and is missing wcs and
                # units. This should likely be represented as an array plus
                # the metadata in the header.
                spec.meta["sami_dust_vector_weights"] = hdu.data

            elif "BIN_MASK" == hdu.header.get("EXTNAME"):
                # This is the bin mask extension, where the value of each
                # pixel indicates the bin to which it belongs. The bin mask
                # is used to construct the binned fluxes and variances in
                # the above two extensions from the default cubes.
                # This is not the same as the aperture spectra mask with the
                # same HDU name.
                spec.meta["sami_bin_mask"] = hdu.data

            else:
                raise NotImplementedError(
                    "Extension is not handled: index {}; name {}".format(
                        i, hdu.header.get("EXTNAME")
                    )
                )

    spectra = SpectrumList(
        spectra_map["combined"] +
        spectra_map["reduced"] +
        spectra_map["unreduced"] +
        spectra_map["sky"]
    )
    return spectra


@data_loader(
    label="SAMI-1d-spec", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_sami_1d_spec, priority=10,
)
def sami_1d_spec_loader(filename):
    spectra_map = {
        "sky": [],
        "combined": [],
        "unreduced": [],
        "reduced": [],
    }
    primary_header = None

    with read_fileobj_or_hdulist(filename) as hdulist:
        for i, hdu in enumerate(hdulist):
            if i == 0:
                # This is the primary extension, and the one with the
                # science data. The header is fairly complete.
                primary_header = hdu.header
                spec = add_single_spectra_to_map(
                    spectra_map,
                    header=primary_header,
                    data=hdu.data,
                    index=None,
                    all_standard_units=True,
                    all_keywords=True,
                    valid_wcs=True,
                )

            elif "VARIANCE" == hdu.header.get("EXTNAME"):
                # This is the variance extension, and is missing wcs and
                # units.
                uncertainty = VarianceUncertainty(
                    hdu.data, unit=spec.flux.unit ** 2
                )
                spec.uncertainty = uncertainty

            elif "BIN_MASK" == hdu.header.get("EXTNAME"):
                # Contains the bin mask used to construct the aperture
                # spectra. A 1 indicates a spaxel was included in the
                # aperture, a 0 indicates a spaxel was not included.
                spec.meta["sami_aperture_spectra_mask"] = hdu.data

            else:
                raise NotImplementedError(
                    "Extension is not handled: index {}; name {}".format(
                        i, hdu.header.get("EXTNAME")
                    )
                )

    spectra = SpectrumList(
        spectra_map["combined"] +
        spectra_map["reduced"] +
        spectra_map["unreduced"] +
        spectra_map["sky"]
    )
    return spectra



from abc import ABC, abstractmethod

import numpy as np
from astropy.nddata import VarianceUncertainty, InverseVariance
from astropy.units import Quantity
from scipy.interpolate import CubicSpline

from ..spectra import Spectrum, SpectralAxis

__all__ = ['ResamplerBase', 'FluxConservingResampler',
           'LinearInterpolatedResampler', 'SplineInterpolatedResampler']


class ResamplerBase(ABC):
    """
    Base class for resample classes.  The algorithms and needs for difference
    resamples will vary quite a bit, so this class is relatively sparse.

    Parameters
    ----------
    extrapolation_treatment : str
        What to do when resampling off the edge of the spectrum.  Can be
        ``'nan_fill'`` to have points beyond the edges by set to NaN,
        ``'zero_fill'`` to set thoe points to zero, or ``'truncate'`` to
        truncate any non-overlapping bins of the spectrum.
    """
    def __init__(self, extrapolation_treatment='nan_fill'):
        if extrapolation_treatment not in ('nan_fill', 'zero_fill', 'truncate'):
            raise ValueError('invalid extrapolation_treatment value: ' + str(extrapolation_treatment))
        self.extrapolation_treatment = extrapolation_treatment

    def __call__(self, orig_spectrum, fin_spec_axis):
        """
        Return the resulting `~specutils.Spectrum` of the resampling.
        """
        return self.resample1d(orig_spectrum, fin_spec_axis)

    @abstractmethod
    def resample1d(self, orig_spectrum, fin_spec_axis):
        """
        Workhorse method that will return the resampled Spectrum
        object.
        """
        return NotImplemented


class FluxConservingResampler(ResamplerBase):

    """
    This resampling algorithm conserves overall integrated flux (as opposed to
    flux density).
    Algorithm based on the equations documented in the following paper:
    https://ui.adsabs.harvard.edu/abs/2017arXiv170505165C/abstract

    Parameters
    ----------
    extrapolation_treatment : str
        What to do when resampling off the edge of the spectrum.  Can be
        ``'nan_fill'`` to have points beyond the edges by set to NaN,
        ``'zero_fill'`` to set those points to zero, or ``'truncate'`` to
        truncate any non-overlapping bins of the spectrum.

    Examples
    --------

    To resample an input spectrum to a user specified spectral grid using
    a flux conserving algorithm:

    >>> import numpy as np
    >>> import astropy.units as u
    >>> from specutils import Spectrum
    >>> from specutils.manipulation import FluxConservingResampler
    >>> input_spectra = Spectrum(
    ...     flux=np.array([1, 3, 7, 6, 20]) * u.mJy,
    ...     spectral_axis=np.array([2, 4, 12, 16, 20]) * u.nm)
    >>> resample_grid = [1, 5, 9, 13, 14, 17, 21, 22, 23]  *u.nm
    >>> fluxc_resample = FluxConservingResampler()
    >>> fluxc_resample(input_spectra, resample_grid)  # doctest: +FLOAT_CMP
    <Spectrum(flux=<Quantity [ 1.  ,  3.  ,  6.  ,  7.  ,  6.25, 10.  , 20.  ,   nan,   nan] mJy> (shape=(9,), mean=7.60714 mJy); spectral_axis=<SpectralAxis [ 1.  5.  9. ... 21. 22. 23.] nm> (length=9))>

    """

    def _fluxc_resample(self, input_bin_centers, output_bin_centers,
                        input_bin_fluxes, errs):
        """
        Resample ``input_bin_fluxes`` and (optionally) ``errs`` from
        ``input_bin_centers`` to ``output_bin_centers``.

        Parameters
        ----------
        input_bin_centers : `~specutils.SpectralAxis`
            `~specutils.SpectralAxis` object, with input bin centers.
        output_bin_centers : `~specutils.SpectralAxis`
            `~specutils.SpectralAxis` object, with input bin centers.
        input_bin_fluxes : Quantity
            Quantity array of flux values.
        errs : `~astropy.nddata.Variance` object, or None
            Variance array of errors corresponding to input bin fluxes. If None,
            error resampling is not performed.

        Returns
        -------
       (output_fluxes, output_errs)
            A tuple containing plain numpy arrays of the resampled fluxes and
            errors (if available).
        """

        fill_val = np.nan  # bin_edges=nan_fill case
        if self.extrapolation_treatment == 'zero_fill':
            fill_val = 0

        # get bin edges from centers
        input_bin_edges = input_bin_centers.bin_edges.value
        output_bin_edges = output_bin_centers.bin_edges.value

        # create array of output fluxes and errors to be returned
        output_fluxes = np.zeros(shape=len(output_bin_centers)) * input_bin_fluxes.unit
        output_errs = None
        if errs is not None:
            output_errs = np.zeros(shape=len(output_bin_centers))

        # first, figure out what output bins cover wavelengths outside the span of
        # input bins. these bins should have fluxes set to nan (or whatever the
        # fill val is.) and can be skipped
        min_idx = 0
        max_idx = None

        low_out_of_range = np.where(output_bin_edges <= input_bin_edges[0])[0]
        if len(low_out_of_range) > 0:  # if any bins below wavelength range
            min_idx = low_out_of_range[-1]  # This doesn't need +1 because bin_edges has len+1 compared to output_fluxes
            output_fluxes[:min_idx] = fill_val
            if errs is not None:
                output_errs[:min_idx] = fill_val

        high_out_of_range = np.where(output_bin_edges > input_bin_edges[-1])[0]
        if len(high_out_of_range) > 0:
            max_idx = high_out_of_range[0] - 1
            output_fluxes[max_idx:] = fill_val
            if errs is not None:
                output_errs[max_idx:] = fill_val

        clipped_output_centers = output_bin_centers[min_idx:max_idx]

        # find the index of the first input bin that intersects the first
        # in-range output bin.
        first_output_edge = output_bin_edges[min_idx]

        idx_last_overlapping_bin = np.where(input_bin_edges[1:] > first_output_edge)[0][0]

        # iterate over each output bin in wavelength range of input bins
        for i, output_bin in enumerate(clipped_output_centers):

            i = i + min_idx  # index in orig, unclipped array

            bin_start, bin_stop = output_bin_edges[i], output_bin_edges[i+1]

            # the first at least partially overlapping bin was determined in the
            # last iteration (or by the initial clipping, if i=0)
            first_bin = idx_last_overlapping_bin

            # keep checking bins, starting at the one after we know overlaps first,
            # and stop when the back edge of an input bin overlaps the front
            # edge of this output bin.
            while input_bin_edges[idx_last_overlapping_bin + 1] < bin_stop:
                idx_last_overlapping_bin += 1

            # if the front edge of the last overlapping bin terminates in this
            # output bin, don't check it next time
            final_bin = idx_last_overlapping_bin

            if input_bin_edges[idx_last_overlapping_bin + 1] <= bin_stop:
                idx_last_overlapping_bin = idx_last_overlapping_bin + 1

            # now, calculate fluxes and errors

            # if only one input bin covers this output bin
            # flux_j=p_ij*w_i*f_i/p_ij*w_i = f_i - f_j=f_i, err_j=err_i
            if final_bin == first_bin:
                output_fluxes[i] = input_bin_fluxes[first_bin]
                if errs is not None:
                    output_errs[i] = errs[first_bin]

            # otherwise, figure out the contribution from each overlapping
            # input bin to calculate the final flux in the output bin.
            else:
                # the first edges of each overlapping input bin
                first_edges = input_bin_edges[first_bin:final_bin+1]
                # the final edges of each overlapping input bin
                second_edges = input_bin_edges[first_bin+1:final_bin+2]

                # to calculate the overlap area, of input on output, we
                # want to only deal with input bin's leading edges if they are
                # inside the output bin. otherwise, they are bounded by the
                # output bin edges. temporarily set the last edges to the output
                # bin bounds and then reset them at the end
                first_edges_orig_first = first_edges[0]
                first_edges[0] = bin_start
                second_edges_orig_last = second_edges[-1]
                second_edges[-1] = bin_stop

                p_ij = second_edges - first_edges

                # reset back
                first_edges[0] = first_edges_orig_first
                second_edges[-1] = second_edges_orig_last

                sum_pij = np.sum(p_ij)

                final_flux = (np.sum(input_bin_fluxes[first_bin:final_bin+1] * p_ij)) / sum_pij
                output_fluxes[i] = final_flux

                if errs is not None:
                    final_err = np.sum((errs[first_bin:final_bin+1] * p_ij) ** 2) / (sum_pij * sum_pij)
                    output_errs[i] = np.sqrt(final_err)

        if errs is not None:
            output_errs = InverseVariance(np.reciprocal(output_errs))

        return (output_fluxes, output_errs)

    def resample1d(self, orig_spectrum, fin_spec_axis):
        """
        Create a re-sampling matrix to be used in re-sampling spectra in a way
        that conserves flux. If an uncertainty is present in the input spectra
        it will be propagated through to the final resampled output spectra
        as an InverseVariance uncertainty.

        Parameters
        ----------
        orig_spectrum : `~specutils.Spectrum`
            The original 1D spectrum.
        fin_spec_axis :  Quantity
            The desired spectral axis array.

        Returns
        -------
        resampled_spectrum : `~specutils.Spectrum`
            An output spectrum containing the resampled `~specutils.Spectrum`
        """

        if isinstance(fin_spec_axis, Quantity):
            if orig_spectrum.spectral_axis.unit != fin_spec_axis.unit:
                raise ValueError("Original spectrum spectral axis grid and new"
                                 "spectral axis grid must have the same units.")

        if not isinstance(fin_spec_axis, SpectralAxis):
            fin_spec_axis = SpectralAxis(fin_spec_axis)

        # Get provided uncertainty into variance
        if orig_spectrum.uncertainty is not None:
            pixel_uncer = orig_spectrum.uncertainty.represent_as(VarianceUncertainty).array
        else:
            pixel_uncer = None

        # convert unit
        orig_axis_in_fin = orig_spectrum.spectral_axis.to(fin_spec_axis.unit)

        # handle multi dimensional flux inputs
        if orig_spectrum.flux.ndim >= 2:

            # the output fluxes and errs should have the same shape as the input
            # except for the last axis, which should be the size of the new
            # spectral axis
            new_shape = tuple(list(orig_spectrum.shape[0:-1]) +
                              [len(fin_spec_axis)])

            # make output matricies
            output_fluxes = np.zeros(shape=new_shape)
            output_errs = np.zeros(shape=new_shape)

            for index, row in np.ndenumerate(orig_spectrum.flux[..., 0]):

                orig_fluxes = orig_spectrum.flux[index]
                orig_uncer = pixel_uncer[index]

                new_f, new_e = self._fluxc_resample(input_bin_centers=orig_axis_in_fin,
                                                    output_bin_centers=fin_spec_axis,
                                                    input_bin_fluxes=orig_fluxes,
                                                    errs=orig_uncer)
                output_fluxes[index] = new_f
                output_errs[index] = new_e.array

            new_errs = InverseVariance(output_errs)

        else:
            # calculate new fluxes and errors
            output_fluxes, new_errs = self._fluxc_resample(input_bin_centers=orig_axis_in_fin,
                                                           output_bin_centers=fin_spec_axis,
                                                           input_bin_fluxes=orig_spectrum.flux,
                                                           errs=pixel_uncer)

        output_fluxes = output_fluxes << orig_spectrum.flux.unit
        fin_spec_axis = np.array(fin_spec_axis) << orig_spectrum.spectral_axis.unit

        if self.extrapolation_treatment == 'truncate':
            fin_spec_axis = fin_spec_axis[np.where(~np.isnan(output_fluxes))]
            if new_errs is not None:
                new_errs = new_errs[np.where(~np.isnan(output_fluxes))]
            output_fluxes = output_fluxes[np.where(~np.isnan(output_fluxes))]

        resampled_spectrum = Spectrum(flux=output_fluxes,
                                        spectral_axis=fin_spec_axis,
                                        uncertainty=new_errs,
                                        spectral_axis_index = orig_spectrum.spectral_axis_index)

        return resampled_spectrum


class LinearInterpolatedResampler(ResamplerBase):
    """
    Resample a spectrum onto a new ``spectral_axis`` using linear interpolation.

    Parameters
    ----------
    extrapolation_treatment : str
        What to do when resampling off the edge of the spectrum.  Can be
        ``'nan_fill'`` to have points beyond the edges by set to NaN,
        ``'zero_fill'`` to set those points to zero, or ``'truncate'`` to
        truncate any non-overlapping bins of the spectrum.

    Examples
    --------

    To resample an input spectrum to a user specified dispersion grid using
    linear interpolation:

    >>> import numpy as np
    >>> import astropy.units as u
    >>> from specutils import Spectrum
    >>> from specutils.manipulation import LinearInterpolatedResampler
    >>> input_spectra = Spectrum(
    ...     flux=np.array([1, 3, 7, 6, 20]) * u.mJy,
    ...     spectral_axis=np.array([2, 4, 12, 16, 20]) * u.nm)
    >>> resample_grid = [1, 5, 9, 13, 14, 17, 21, 22, 23] * u.nm
    >>> fluxc_resample = LinearInterpolatedResampler()
    >>> fluxc_resample(input_spectra, resample_grid)  # doctest: +FLOAT_CMP
    <Spectrum(flux=<Quantity [ nan, 3.5 , 5.5 , 6.75, 6.5 , 9.5 ,  nan,  nan,  nan] mJy> (shape=(9,), mean=6.35000 mJy); spectral_axis=<SpectralAxis [ 1.  5.  9. ... 21. 22. 23.] nm> (length=9))>

    """
    def __init__(self, extrapolation_treatment='nan_fill'):
        super().__init__(extrapolation_treatment)

    def resample1d(self, orig_spectrum, fin_spec_axis):
        """
        Call interpolation, repackage new spectra


        Parameters
        ----------
        orig_spectrum : `~specutils.Spectrum`
            The original 1D spectrum.
        fin_spec_axis : ndarray
            The desired spectral axis array.

        Returns
        -------
        resample_spectrum : `~specutils.Spectrum`
            An output spectrum containing the resampled `~specutils.Spectrum`
        """

        fill_val = np.nan  # bin_edges=nan_fill case
        if self.extrapolation_treatment == 'zero_fill':
            fill_val = 0

        orig_axis_in_fin = orig_spectrum.spectral_axis.to(fin_spec_axis.unit)

        out_flux_arr = np.interp(fin_spec_axis.value, orig_axis_in_fin.value,
                                 orig_spectrum.flux.value, left=fill_val, right=fill_val)
        out_flux = Quantity(out_flux_arr, unit=orig_spectrum.flux.unit)

        new_unc = None
        if orig_spectrum.uncertainty is not None:
            out_unc_arr = np.interp(fin_spec_axis.value, orig_axis_in_fin.value,
                                    orig_spectrum.uncertainty.array,
                                    left=fill_val, right=fill_val)
            new_unc = orig_spectrum.uncertainty.__class__(array=out_unc_arr,
                                                          unit=orig_spectrum.unit)

        if self.extrapolation_treatment == 'truncate':
            fin_spec_axis = fin_spec_axis[np.where(~np.isnan(out_flux))]
            if new_unc is not None:
                new_unc = new_unc[np.where(~np.isnan(out_flux))]
            out_flux = out_flux[np.where(~np.isnan(out_flux))]

        return Spectrum(spectral_axis=fin_spec_axis,
                          flux=out_flux,
                          uncertainty=new_unc,
                          spectral_axis_index = orig_spectrum.spectral_axis_index)


class SplineInterpolatedResampler(ResamplerBase):
    """
    This resample algorithim uses a cubic spline interpolator. Any uncertainty
    is also interpolated using an identical spline.


    Parameters
    ----------
    extrapolation_treatment : str
        What to do when resampling off the edge of the spectrum.  Can be
        ``'nan_fill'`` to have points beyond the edges by set to NaN,
        ``'zero_fill'`` to set those points to zero, or ``'truncate'`` to
        truncate any non-overlapping bins of the spectrum. Any other value will
        have the spline interpolate beyond the edges of the original data.

    Examples
    --------

    To resample an input spectrum to a user specified spectral axis grid using
    a cubic spline interpolator:

    >>> import numpy as np
    >>> import astropy.units as u
    >>> from specutils import Spectrum
    >>> from specutils.manipulation import SplineInterpolatedResampler
    >>> input_spectra = Spectrum(
    ...     flux=np.array([1, 3, 7, 6, 20]) * u.mJy,
    ...     spectral_axis=np.array([2, 4, 12, 16, 20]) * u.nm)
    >>> resample_grid = [1, 5, 9, 13, 14, 17, 21, 22, 23] * u.nm
    >>> fluxc_resample = SplineInterpolatedResampler()
    >>> fluxc_resample(input_spectra, resample_grid)  # doctest: +FLOAT_CMP
    <Spectrum(flux=<Quantity [       nan, 3.98808594, 6.94042969, 6.45869141, 5.89921875,
               7.29736328,        nan,        nan,        nan] mJy> (shape=(9,), mean=6.11676 mJy); spectral_axis=<SpectralAxis [ 1.  5.  9. ... 21. 22. 23.] nm> (length=9))>

    """
    def __init__(self, extrapolation_treatment='nan_fill'):
        super().__init__(extrapolation_treatment)

    def resample1d(self, orig_spectrum, fin_spec_axis):
        """
        Call interpolation, repackage new spectra


        Parameters
        ----------
        orig_spectrum : `~specutils.Spectrum`
            The original 1D spectrum.
        fin_spec_axis : Quantity
            The desired spectral axis array.

        Returns
        -------
        resample_spectrum : `~specutils.Spectrum`
            An output spectrum containing the resampled `~specutils.Spectrum`
        """
        orig_axis_in_new = orig_spectrum.spectral_axis.to(fin_spec_axis.unit)
        flux_spline = CubicSpline(orig_axis_in_new.value, orig_spectrum.flux.value,
                                  extrapolate=self.extrapolation_treatment not in ('nan_fill',
                                                                                   'zero_fill',
                                                                                   'truncate'))
        out_flux_val = flux_spline(fin_spec_axis.value)

        new_unc = None
        if orig_spectrum.uncertainty is not None:
            unc_spline = CubicSpline(orig_axis_in_new.value, orig_spectrum.uncertainty.array,
                                     extrapolate=self.extrapolation_treatment not in ('nan_fill',
                                                                                      'zero_fill',
                                                                                      'truncate'))
            out_unc_val = unc_spline(fin_spec_axis.value)
            new_unc = orig_spectrum.uncertainty.__class__(array=out_unc_val, unit=orig_spectrum.unit)

        fill_val = np.nan
        if self.extrapolation_treatment == 'zero_fill':
            fill_val = 0

        orig_edges = orig_axis_in_new.bin_edges
        off_edges = (fin_spec_axis < np.min(orig_edges)) | (np.max(orig_edges) < fin_spec_axis)
        out_flux_val[off_edges] = fill_val
        if new_unc is not None:
            new_unc.array[off_edges] = fill_val

        if self.extrapolation_treatment == 'truncate':
            fin_spec_axis = fin_spec_axis[np.where(~np.isnan(out_flux_val))]
            if new_unc is not None:
                new_unc = new_unc[np.where(~np.isnan(out_flux_val))]
            out_flux_val = out_flux_val[np.where(~np.isnan(out_flux_val))]

        return Spectrum(spectral_axis=fin_spec_axis,
                          flux=out_flux_val*orig_spectrum.flux.unit,
                          uncertainty=new_unc,
                          spectral_axis_index = orig_spectrum.spectral_axis_index)



"""
A module for analysis tools focused on determining the moment of
spectral features.
"""

import numpy as np
from ..manipulation import extract_region
from ..spectra import SpectrumCollection
from .utils import computation_wrapper


__all__ = ['moment']


def moment(spectrum, regions=None, order=0, axis='spectral'):
    """
    Estimate the moment of the spectrum.


    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum.Spectrum`
        The spectrum object over which the width will be calculated.

    regions: `~specutils.SpectralRegion` or list of `~specutils.SpectralRegion`
        Region within the spectrum to calculate the gaussian sigma width. If
        regions is `None`, computation is performed over entire spectrum.

    order : int
        The order of the moment to be calculated. Default=0

    axis : int, str
        Axis along which a moment is calculated. Default='spectral', which computes
        along the spectral axis.


    Returns
    -------
    moment: `float` or list (based on region input)
        Moment of the spectrum. Returns None if (order < 0 or None)

    """
    if isinstance(spectrum, SpectrumCollection):
        return [computation_wrapper(_compute_moment, spec, regions,order=order, axis=axis)
                for spec in spectrum]
    return computation_wrapper(_compute_moment, spectrum, regions,
                               order=order, axis=axis)


def _compute_moment(spectrum, regions=None, order=0, axis='spectral'):
    """
    This is a helper function for the above `moment()` method.
    """
    if int(order) != order or order < 0:
        raise ValueError("Order must be a positive integer.")

    if axis == "spectral":
        if isinstance(spectrum, SpectrumCollection):
            axes = [spec.spectral_axis_index for spec in spectrum]
            if not np.all([x==axes[0] for x in axes]):
                raise ValueError("All spectra in SpectrumCollection must have the same "
                                 "spectral_axis_index for simultaneous moment calculation.")
            # SpectumCollection adds a leading axis when it stacks the spectra.
            axis = axes[0]+1

        else:
            axis = spectrum.spectral_axis_index

    if regions is not None:
        calc_spectrum = extract_region(spectrum, regions)
    else:
        calc_spectrum = spectrum

    # Ignore masks for now. This should be fully addressed when
    # specutils gets revamped to handle multi-dimensional masks.
    flux = calc_spectrum.flux

    spectral_axis = calc_spectrum.spectral_axis
    # If axis is not the spectral axis, we simply use pixel values for dx and dispersion
    if (axis != spectrum.spectral_axis_index and
            not (axis == -1 and spectrum.spectral_axis_index == spectrum.flux.ndim-1)):
        dx = np.ones(flux.shape[axis])
        dispersion = np.arange(flux.shape[axis]) + 1
    else:
        dx = np.abs(np.diff(spectral_axis.bin_edges))
        dispersion = spectral_axis

    # We now have to account for the desired axis being anywhere, not always last
    if len(flux.shape) > len(spectral_axis.shape):
        for i in range(flux.ndim):
            if i != axis:
                dx = np.expand_dims(dx, i)
                dx = np.repeat(dx, flux.shape[i], i)
                dispersion = np.expand_dims(dispersion, i)
                dispersion = np.repeat(dispersion, flux.shape[i], i)

    m0 = np.sum(flux * dx, axis=axis)
    if order == 0:
        return m0

    if order == 1:
        return np.sum(flux * dispersion * dx, axis=axis) / m0

    if order > 1:

        # By setting keepdims to True, the axes which are reduced are
        # left in the result as dimensions with size one. This means
        # that we can broadcast m1 correctly against dispersion.
        m1 = (np.sum(flux * dispersion * dx, axis=axis, keepdims=True)
              / np.sum(flux * dx, axis=axis, keepdims=True))

        return np.sum(flux * dx * (dispersion - m1) ** order, axis=axis) / m0



try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
    ASTROPY_HEADER = True
except ImportError:
    ASTROPY_HEADER = False


# Repeat this from specutils/conftest.py so tox picks it up.
def pytest_configure(config):

    if ASTROPY_HEADER:
        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the tests.
        PYTEST_HEADER_MODULES.pop('Pandas', None)
        PYTEST_HEADER_MODULES.pop('h5py', None)
        PYTEST_HEADER_MODULES['astropy'] = 'astropy'
        PYTEST_HEADER_MODULES['gwcs'] = 'gwcs'
        PYTEST_HEADER_MODULES['asdf'] = 'asdf'
        PYTEST_HEADER_MODULES['asdf-astropy'] = 'asdf_astropy'
        PYTEST_HEADER_MODULES['stdatamodels'] = 'stdatamodels'
        PYTEST_HEADER_MODULES['ndcube'] = 'ndcube'
        PYTEST_HEADER_MODULES['spectral-cube'] = 'spectral_cube'

        from specutils import __version__
        TESTED_VERSIONS['specutils'] = __version__



from math import floor, ceil  # faster than int(np.floor/ceil(float))

import numpy as np

from astropy import units as u
from ..spectra import Spectrum, SpectralRegion

__all__ = ['extract_region', 'extract_bounding_spectral_region', 'spectral_slab']


def _edge_value_to_pixel(edge_value, spectrum, order, side):
    spectral_axis = spectrum.spectral_axis

    if order == 'ascending':
        index = np.searchsorted(spectral_axis, edge_value, side=side)
        return index

    elif order == 'descending':

        if side == 'left':
            opposite_side = 'right'
        else:
            opposite_side = 'left'

        index = len(spectral_axis) - np.searchsorted(spectral_axis[::-1], edge_value, side=opposite_side)
        return index


def _subregion_to_edge_pixels(subregion, spectrum):
    """
    Calculate and return the left and right indices defined
    by the lower and upper bounds and based on the input
    `~specutils.spectra.spectrum.Spectrum`. The left and right indices will
    be returned.

    Parameters
    ----------
    spectrum: `~specutils.spectra.spectrum.Spectrum`
        The spectrum object from which the region will be extracted.

    Returns
    -------
    left_index, right_index: int, int
        Left and right indices defined by the lower and upper bounds.

    """
    spectral_axis = spectrum.spectral_axis
    if spectral_axis[-1] > spectral_axis[0]:
        order = "ascending"
        left_func = min
        right_func = max
    else:
        order = "descending"
        left_func = max
        right_func = min

    # Left/lower side of sub region
    if subregion[0].unit.is_equivalent(u.pix):
        # Pixel handling assumes always ascending
        if not spectral_axis.unit.is_equivalent(u.pix):
            left_index = floor(subregion[0].value)
        else:
            # If the lower bound is larger than the largest value, immediately return nothing
            # Assuming ascending, both bounds are "out of bounds"
            if subregion[0] > spectral_axis[-1]:
                return None, None
            else:
                # Get index of closest value
                # See https://stackoverflow.com/a/26026189 if performance becomes an issue
                left_index = np.nanargmin((np.abs(spectral_axis - subregion[0])))
                # Ensure index is inclusive of region bounds
                if (spectral_axis[left_index] > subregion[0]) and (left_index >= 1):
                    left_index -= 1
    else:
        # Convert lower value to spectrum spectral_axis units
        left_reg_in_spec_unit = left_func(subregion).to(spectral_axis.unit,
                                                        u.spectral())
        left_index = _edge_value_to_pixel(left_reg_in_spec_unit, spectrum, order, "left")

    # Right/upper side of sub region
    if subregion[1].unit.is_equivalent(u.pix):
        # Pixel handling assumes always ascending
        if not spectral_axis.unit.is_equivalent(u.pix):
            right_index = ceil(subregion[1].value)
        else:
            # If upper bound is smaller than smallest value, immediately return nothing
            # Assuming ascending, both bounds are "out of bounds"
            if subregion[1] < spectral_axis[0]:
                return None, None
            else:
                # Get index of closest value
                # See https://stackoverflow.com/a/26026189 if performance becomes an issue
                right_index = np.nanargmin((np.abs(spectral_axis - subregion[1])))
                # Ensure index is inclusive of region bounds
                if (spectral_axis[right_index] < subregion[1]) and (right_index < len(spectral_axis)):
                    right_index += 1
    else:
        # Convert upper value to spectrum spectral_axis units
        right_reg_in_spec_unit = right_func(subregion).to(spectral_axis.unit,
                                                 u.spectral())

        right_index = _edge_value_to_pixel(right_reg_in_spec_unit, spectrum, order, "right")

    # If the spectrum is in wavelength and region is in Hz (for example), these still might be reversed
    if left_index < right_index:
        return left_index, right_index
    else:
        return right_index, left_index


def extract_region(spectrum, region, return_single_spectrum=False):
    """
    Extract a region from the input `~specutils.Spectrum`
    defined by the lower and upper bounds defined by the ``region``
    instance.  The extracted region will be returned as a new
    `~specutils.Spectrum`.

    Parameters
    ----------
    spectrum: `~specutils.Spectrum`
        The spectrum object from which the region will be extracted.

    region: `~specutils.SpectralRegion`
        The spectral region to extract from the original spectrum.

    return_single_spectrum: `bool`
        If ``region`` has multiple sections, whether to return a single spectrum
        instead of multiple `~specutils.Spectrum` objects.  The returned spectrum
        will be a unique, concatenated, spectrum of all sub-regions.

    Returns
    -------
    spectrum: `~specutils.Spectrum` or list of `~specutils.Spectrum`
        Excised spectrum, or list of spectra if the input region contained multiple
        subregions and ``return_single_spectrum`` is `False`.

    Notes
    -----
    The region extracted is a discrete subset of the input spectrum. No interpolation is done
    on the left and right side of the spectrum.

    The region is assumed to be a closed interval (as opposed to Python which is open
    on the upper end).  For example:

        Given:
           A ``spectrum`` with spectral_axis of ``[0.1, 0.2, 0.3, 0.4, 0.5, 0.6]*u.um``.

           A ``region`` defined as ``SpectralRegion(0.2*u.um, 0.5*u.um)``

        And we calculate ``sub_spectrum = extract_region(spectrum, region)``, then the ``sub_spectrum``
        spectral axis will be ``[0.2, 0.3, 0.4, 0.5] * u.um``.

    If the ``region`` does not overlap with the ``spectrum`` then an empty Spectrum object
    will be returned.

    """
    extracted_spectrum = []
    for subregion in region._subregions:
        left_index, right_index = _subregion_to_edge_pixels(subregion, spectrum)

        # If both indices are out of bounds then return an empty spectrum
        if left_index == right_index:
            empty_spectrum = Spectrum(spectral_axis=[]*spectrum.spectral_axis.unit,
                                        flux=[]*spectrum.flux.unit)
            extracted_spectrum.append(empty_spectrum)
        else:
            slices = [slice(None),] * len(spectrum.shape)
            slices[spectrum.spectral_axis_index] = slice(left_index, right_index)
            if len(slices) == 1:
                slices = slices[0]
            else:
                slices = tuple(slices)
            extracted_spectrum.append(spectrum[slices])

    # If there is only one subregion in the region then we will
    # just return a spectrum.
    if len(region) == 1:
        extracted_spectrum = extracted_spectrum[0]

    # Otherwise, if requested to return a single spectrum, we need to combine
    # the Spectrum objects in extracted_spectrum and return a single object.
    elif return_single_spectrum:
        concat_keys = ['flux', 'uncertainty', 'mask']  # spectral_axis handled manually
        copy_keys = ['velocity_convention', 'rest_value', 'meta']

        # NOTE: WCS is intentionally dropped, which will then fallback on lookup

        def _get_joined_value(sps, key, unique_inds=None):
            if key == 'uncertainty':
                # uncertainty cannot be appended directly as its an object,
                # not an array so instead we'll take a copy of the first entry
                # and overwrite the internal array with an appended array
                uncert = sps[0].uncertainty
                if uncert is None:
                    return None
                uncert._array = np.concatenate([sp.uncertainty._array for sp in sps])
                return uncert[unique_inds] if unique_inds is not None else uncert
            elif key in concat_keys or key == 'spectral_axis':
                if getattr(sps[0], key) is None:
                    return None
                concat_arr = np.concatenate([getattr(sp, key) for sp in sps])
                return concat_arr[unique_inds] if unique_inds is not None else concat_arr
            else:
                # all were extracted from the same input spectrum, so we don't
                # need to make sure the properties match
                return getattr(sps[0], key)

        # we'll need to account for removing overlapped regions in the spectral axis,
        # so we'll concatenate that first and track the unique indices
        spectral_axis = _get_joined_value(extracted_spectrum, 'spectral_axis')
        spectral_axis_unique, unique_inds = np.unique(spectral_axis, return_index=True)
        return Spectrum(spectral_axis=spectral_axis_unique,
                          **{key: _get_joined_value(extracted_spectrum, key, unique_inds)
                             for key in concat_keys+copy_keys})

    return extracted_spectrum


def spectral_slab(spectrum, lower, upper):
    """
    Extract a slab from the input `~specutils.Spectrum`
    defined by the lower and upper bounds defined by the ``region``
    instance.  The extracted region will be returned as a new
    `~specutils.Spectrum`.

    Parameters
    ----------
    spectrum: `~specutils.Spectrum`
        The spectrum object from which the region will be extracted.

    lower, upper: `~astropy.units.Quantity`
        The lower and upper bounds of the region to extract
        from the original spectrum.

    Returns
    -------
    spectrum: `~specutils.Spectrum` or list of `~specutils.Spectrum`
        Excised spectrum, or list of spectra if the input region contained multiple
        subregions.

    Notes
    -----
    This is for now just a proxy for function `extract_region`, to ease the
    transition from spectral-cube.

    """
    region = SpectralRegion(lower, upper)

    return extract_region(spectrum, region)


def extract_bounding_spectral_region(spectrum, region):
    """
    Extract the entire bounding region that encompasses all sub-regions
    contained in a multi-sub-region instance of `~specutils.SpectralRegion`.

    In case only one sub-region exists, this method is equivalent to
    `extract_region`.

    Parameters
    ----------
    spectrum: `~specutils.Spectrum`
        The spectrum object from which the region will be extracted.

    region: `~specutils.SpectralRegion`
        The spectral region to extract from the original spectrum, comprised
        of one or more sub-regions.

    Returns
    -------
    spectrum: `~specutils.Spectrum`
        Excised spectrum from the bounding region defined by the set of
        sub-regions in the input ``region`` instance.

    """
    # If there is only one subregion in the region then we will
    # just return a spectrum.
    if len(region) == 1:
        return extract_region(spectrum, region)

    # Look for indices that bound the entire set of sub-regions.
    min_list = [min(sr) for sr in region._subregions]
    max_list = [max(sr) for sr in region._subregions]

    single_region = SpectralRegion(min(min_list), max(max_list))

    return extract_region(spectrum, single_region)



"""This test module exists because Jdaviz wanted to put pixel area
in the flux unit. Some code copied over from original Jdaviz
implementation.

"""
import numpy as np
import pytest
from astropy import units as u
from astropy.nddata import StdDevUncertainty
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import WCS
from numpy.testing import assert_array_equal

from specutils import Spectrum, SpectralAxis

PIX2 = u.pix * u.pix


def _eqv_flux_to_sb_pixel():
    """This allows conversion between flux and flux-per-square-pixel
    surface brightness, e.g., MJy and MJy/PIX2.
    """

    # generate an equivalency for each flux type that would need
    # another equivalency for converting to/from
    flux_units = [u.MJy, u.erg / (u.s * u.cm**2 * u.Angstrom),
                  u.ph / (u.Angstrom * u.s * u.cm**2),
                  u.ph / (u.Hz * u.s * u.cm**2)]
    return [(flux_unit, flux_unit / PIX2, lambda x: x, lambda x: x)
            for flux_unit in flux_units]


# The original Jdaviz implementation we are replacing with native
# specutils functionality.
def convert_spectrum1d_from_flux_to_flux_per_pixel(spectrum):
    """Converts a Spectrum object's flux units to flux per square pixel.

    This function takes a `specutils.Spectrum` object with flux units and converts the
    flux (and optionally, uncertainty) to a surface brightness per square pixel
    (e.g., from Jy to Jy/pix**2). This is done by updating the units of spectrum.flux
    and (if present) spectrum.uncertainty, and creating a new `specutils.Spectrum`
    object with the modified flux and uncertainty.

    Parameters
    ----------
    spectrum : Spectrum
        A `specutils.Spectrum` object containing flux data, which is assumed to be in
        flux units without any angular component in the denominator.

    Returns
    -------
    Spectrum
        A new `specutils.Spectrum` object with flux and uncertainty (if present)
        converted to units of flux per square pixel.

    """
    # convert flux, which is always populated
    flux = getattr(spectrum, 'flux')
    flux = flux / PIX2

    # and uncerts, if present
    uncerts = getattr(spectrum, 'uncertainty')
    if uncerts is not None:
        # enforce common uncert type.
        uncerts = uncerts.represent_as(StdDevUncertainty)
        uncerts = StdDevUncertainty(uncerts.quantity / PIX2)

    # create a new spectrum 1d with all the info from the input spectrum 1d,
    # and the flux / uncerts converted from flux to SB per square pixel

    # if there is a spectral axis that is a SpectralAxis, you cant also set
    # redshift or radial_velocity
    spectral_axis = getattr(spectrum, 'spectral_axis', None)
    if spectral_axis is not None:
        if isinstance(spectral_axis, SpectralAxis):
            redshift = None
            radial_velocity = None
        else:
            redshift = spectrum.redshift
            radial_velocity = spectrum.radial_velocity

    # initialize new spectrum1d with new flux, uncerts, and all other init parameters
    # from old input spectrum as well as any 'meta'. any more missing information
    # not in init signature that might be present in `spectrum`?
    new_spec1d = Spectrum(flux=flux, uncertainty=uncerts,
                            spectral_axis=spectrum.spectral_axis,
                            mask=spectrum.mask,
                            wcs=spectrum.wcs,
                            velocity_convention=spectrum.velocity_convention,
                            rest_value=spectrum.rest_value, redshift=redshift,
                            radial_velocity=radial_velocity,
                            bin_specification=getattr(spectrum, 'bin_specification', None),
                            meta=spectrum.meta, spectral_axis_index=spectrum.spectral_axis_index)

    return new_spec1d


def assert_dict_equal(d1, d2):
    keys = sorted(d1)
    assert sorted(d2) == keys
    for k in keys:
        v1 = d1[k]
        v2 = d2[k]
        assert v1 == v2


@pytest.mark.parametrize("suppress_conversion", [False, True])
def test_spec_flux_conv_pix2(suppress_conversion):
    meta = {"CTYPE1": "WAVE-LOG", "CTYPE2": "DEC--TAN", "CTYPE3": "RA---TAN",
            "CRVAL1": 4.622e-7, "CRVAL2": 27, "CRVAL3": 205,
            "CDELT1": 8e-11, "CDELT2": 0.0001, "CDELT3": -0.0001,
            "CRPIX1": 0, "CRPIX2": 0, "CRPIX3": 0, "PIXAR_SR": 8e-11}
    w = WCS(meta)
    flux_orig = np.arange(24).reshape((2, 3, 4)) * u.Jy
    uncert_orig = StdDevUncertainty(flux_orig)
    mask_orig = np.zeros(flux_orig.shape, dtype=bool)
    rs_orig = 0.0001 * u.dimensionless_unscaled
    sp_orig = Spectrum(
        flux=flux_orig, uncertainty=uncert_orig, mask=mask_orig, wcs=w,
        redshift=rs_orig, meta=meta)

    # What Jdaviz implemented.
    sp_pix2_jdaviz = convert_spectrum1d_from_flux_to_flux_per_pixel(sp_orig)

    # What we should replace it with using only specutils.
    sp_pix2_specutils = sp_orig.with_flux_unit(
        u.Jy / PIX2, equivalencies=_eqv_flux_to_sb_pixel(),
        suppress_conversion=suppress_conversion)

    # Make sure the two implementations are equivalent.
    assert sp_pix2_specutils.flux.unit == u.Jy / PIX2
    assert sp_pix2_specutils.uncertainty.unit == u.Jy / PIX2
    assert_quantity_allclose(sp_pix2_specutils.flux, sp_pix2_jdaviz.flux)
    assert_quantity_allclose(
        sp_pix2_specutils.uncertainty.quantity, sp_pix2_jdaviz.uncertainty.quantity)
    assert_quantity_allclose(sp_pix2_specutils.spectral_axis, sp_pix2_jdaviz.spectral_axis)
    assert_array_equal(sp_pix2_specutils.mask, sp_pix2_jdaviz.mask)
    assert_dict_equal(sp_pix2_specutils.wcs.to_header(), sp_pix2_jdaviz.wcs.to_header())
    assert_dict_equal(sp_pix2_specutils.meta, sp_pix2_jdaviz.meta)
    assert sp_pix2_specutils.velocity_convention == sp_pix2_jdaviz.velocity_convention
    assert sp_pix2_specutils.rest_value is sp_pix2_jdaviz.rest_value  # None
    assert_quantity_allclose(sp_pix2_specutils.redshift, sp_pix2_jdaviz.redshift)
    assert_quantity_allclose(sp_pix2_specutils.radial_velocity, sp_pix2_jdaviz.radial_velocity)
    if "bin_specification" in sp_orig:
        assert sp_pix2_specutils.bin_specification == sp_pix2_jdaviz.bin_specification



"""
Loader for DESI spectrum files: spectra_ and coadd_.

* spectra_ files contain all observations of the objects in a given region.
* coadd_ files contain one, co-added spectrum of each object in a given region.
  The coaddition is performed across observations, but *not* across the
  three spectrographs_.

The "region" can be a HEALPixel or a tile_.  DESI does not provide spectra for
individual objects in a file-based form.

.. _spectra: https://desidatamodel.readthedocs.io/en/latest/DESI_SPECTRO_REDUX/SPECPROD/healpix/SURVEY/PROGRAM/PIXGROUP/PIXNUM/spectra-SURVEY-PROGRAM-PIXNUM.html
.. _coadd: https://desidatamodel.readthedocs.io/en/latest/DESI_SPECTRO_REDUX/SPECPROD/healpix/SURVEY/PROGRAM/PIXGROUP/PIXNUM/coadd-SURVEY-PROGRAM-PIXNUM.html
.. _spectrographs: https://data.desi.lbl.gov/doc/glossary/#spectrograph
.. _tile: https://data.desi.lbl.gov/doc/glossary/#tile
"""
import warnings

from astropy.table import Table
from astropy.units import Unit
from astropy.nddata import InverseVariance
from astropy.utils.exceptions import AstropyUserWarning

import numpy as np

from ...spectra import Spectrum, SpectrumList
from ..registers import data_loader
from ..parsing_utils import _fits_identify_by_name, read_fileobj_or_hdulist

__all__ = ['spectra_identify', 'coadd_identify',
           'spectra_loader', 'coadd_loader']

_desi_patterns = {'spectra': {'healpix': r'spectra-(cmx|main|special|sv1|sv2|sv3)-(backup|bright|dark|other)-[0-9]+\.fits',
                              'tile': r'spectra-[0-9]-[0-9]+-([14]xsubset[1-6]|lowspeedsubset[1-6]|exp[0-9]{8}|thru[0-9]{8}|[0-9]{8})\.fits'},
                  'coadd': {'healpix': r'coadd-(cmx|main|special|sv1|sv2|sv3)-(backup|bright|dark|other)-[0-9]+\.fits',
                            'tile': r'coadd-[0-9]-[0-9]+-([14]xsubset[1-6]|lowspeedsubset[1-6]|exp[0-9]{8}|thru[0-9]{8}|[0-9]{8})\.fits'}}


def spectra_identify(origin, *args, **kwargs):
    """
    Check whether given input is FITS and appears to be a DESI spectra file.
    This is used for Astropy I/O Registry.
    """
    return _desi_identify('spectra', origin, *args, **kwargs)


def coadd_identify(origin, *args, **kwargs):
    """
    Check whether given input is FITS and appears to be a DESI coadd file.
    This is used for Astropy I/O Registry.
    """
    return _desi_identify('coadd', origin, *args, **kwargs)


def _desi_identify(desitype, origin, *args, **kwargs):
    """This contains the common, low-level code for identifying spectra and coadd files.
    """
    check_healpix = _fits_identify_by_name(origin, *args, pattern=_desi_patterns[desitype]['healpix'])
    check_tile = _fits_identify_by_name(origin, *args, pattern=_desi_patterns[desitype]['tile'])
    if check_healpix or check_tile:
        return True
    else:
        # If the input was a file-like object:
        with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
            check0 = ('FIBERMAP' in hdulist and
                      'SCORES' in hdulist)
            if desitype == 'spectra':
                check1 = 'INFIL000' not in hdulist[0].header
            else:
                check1 = 'INFIL000' in hdulist[0].header
        return check0 and check1


@data_loader(
    label="DESI spectra", identifier=spectra_identify, dtype=SpectrumList, extensions=['fits'],
    priority=10,
)
def spectra_loader(file_obj, **kwargs):
    """Loader for DESI spectra_ files.

    .. _spectra: https://desidatamodel.readthedocs.io/en/latest/DESI_SPECTRO_REDUX/SPECPROD/healpix/SURVEY/PROGRAM/PIXGROUP/PIXNUM/spectra-SURVEY-PROGRAM-PIXNUM.html

    Any keyword arguments are passed to
    :func:`~specutils.io.parsing_utils.read_fileobj_or_hdulist`.

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
          FITS file name, object (provided from name by Astropy I/O Registry),
          or HDUList (as resulting from astropy.io.fits.open()).

    Returns
    -------
    SpectrumList
        Each spectrograph arm, or 'band' is represented as a Spectrum
        object in the SpectrumList.
    """
    return _read_desi(file_obj, **kwargs)


@data_loader(
    label="DESI coadd", identifier=coadd_identify, dtype=SpectrumList, extensions=['fits'],
    priority=10,
)
def coadd_loader(file_obj, **kwargs):
    """Loader for DESI coadd_ files.

    .. _coadd: https://desidatamodel.readthedocs.io/en/latest/DESI_SPECTRO_REDUX/SPECPROD/healpix/SURVEY/PROGRAM/PIXGROUP/PIXNUM/coadd-SURVEY-PROGRAM-PIXNUM.html

    Any keyword arguments are passed to
    :func:`~specutils.io.parsing_utils.read_fileobj_or_hdulist`.

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
          FITS file name, object (provided from name by Astropy I/O Registry),
          or HDUList (as resulting from astropy.io.fits.open()).

    Returns
    -------
    SpectrumList
        Each spectrograph arm, or 'band' is represented as a Spectrum
        object in the SpectrumList.
    """
    return _read_desi(file_obj, **kwargs)


def _read_desi(file_obj, **kwargs):
    """Read DESI data from a FITS file.

    This contains the common, low-level code for reading spectra and coadd files.

    Any keyword arguments are passed to
    :func:`~specutils.io.parsing_utils.read_fileobj_or_hdulist`.

    Parameters
    ----------
    file_obj: str, file-like, or HDUList
          FITS file name, object (provided from name by Astropy I/O Registry),
          or HDUList (as resulting from astropy.io.fits.open()).

    Returns
    -------
    SpectrumList
        Each spectrograph arm, or 'band' is represented as a Spectrum
        object in the SpectrumList.
    """
    expected_hdus = ('PRIMARY', 'FIBERMAP', 'EXP_FIBERMAP',
                     'B_WAVELENGTH', 'B_FLUX', 'B_IVAR', 'B_MASK', 'B_RESOLUTION',
                     'R_WAVELENGTH', 'R_FLUX', 'R_IVAR', 'R_MASK', 'R_RESOLUTION',
                     'Z_WAVELENGTH', 'Z_FLUX', 'Z_IVAR', 'Z_MASK', 'Z_RESOLUTION',
                     'SCORES', 'EXTRA_CATALOG')
    tables = ('FIBERMAP', 'EXP_FIBERMAP', 'SCORES', 'EXTRA_CATALOG')
    sl = SpectrumList()
    meta_zero = dict()
    band_data = dict()
    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        for k, hdu in enumerate(hdulist):
            if 'EXTNAME' in hdu.header:
                extname = hdu.header['EXTNAME']
            else:
                if k == 0:
                    extname = 'PRIMARY'
                else:
                    warnings.warn(f"HDU {k:d} has no EXTNAME and will not be read!",
                                  AstropyUserWarning)
                    extname = 'UNKNOWN'
            if extname not in expected_hdus:
                if extname != 'UNKNOWN':  # We already warned about UNKNOWN above.
                    warnings.warn(f"HDU {k:d}, EXTNAME='{extname}' is unexpected and will not be read!",
                                  AstropyUserWarning)
            elif extname == 'PRIMARY':
                meta_zero['header'] = hdu.header.copy()
            elif extname in tables:
                meta_zero[extname.lower()] = Table(hdu.data, copy=True)
            else:
                foo = extname.split('_')
                band = foo[0].lower()
                datatype = foo[1]
                if band not in band_data:
                    band_data[band] = {'meta': {'band': band}}
                if datatype == 'WAVELENGTH':
                    band_data[band]['spectral_axis'] = hdu.data.copy() * Unit(hdu.header['BUNIT'])
                if datatype == 'FLUX':
                    meta_zero['single'] = (hdu.data.dtype == np.dtype('>f4'))
                    band_data[band]['flux'] = hdu.data.copy() * Unit(hdu.header['BUNIT'])
                if datatype == 'IVAR':
                    band_data[band]['uncertainty'] = InverseVariance(hdu.data.copy() * Unit(hdu.header['BUNIT']))
                if datatype == 'MASK':
                    band_data[band]['meta']['int_mask'] = hdu.data.copy()
                    band_data[band]['mask'] = band_data[band]['meta']['int_mask'] != 0
                if datatype == 'RESOLUTION':
                    band_data[band]['meta']['resolution_data'] = hdu.data.copy()
    meta_zero['bands'] = sorted(band_data.keys())
    for i, band in enumerate(meta_zero['bands']):
        if i == 0:
            for key, value in meta_zero.items():
                band_data[band]['meta'][key] = value
        sl.append(Spectrum(**(band_data[band])))
    return sl



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Specutils: an astropy package for spectroscopy.
"""
import os

from astropy import config as _config

__citation__ = 'https://doi.org/10.5281/zenodo.1421356'

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''

# Allow loading spectrum object from top level module
from .spectra import *  # noqa

# Load the IO functions
from .io.default_loaders import *  # noqa
from .io.registers import _load_user_io  # noqa
_load_user_io()


class Conf(_config.ConfigNamespace):
    """Configuration parameters for specutils."""

    do_continuum_function_check = _config.ConfigItem(
        True,
        'Whether to check the spectrum baseline value is close'
        'to zero. If it is not within ``threshold`` then a warning is raised.'
    )


conf = Conf()

# Clean up namespace
del os
del _config
del _load_user_io



import numpy as np

import astropy.units as u
from astropy.nddata import StdDevUncertainty
from astropy.modeling import models

from astropy.tests.helper import assert_quantity_allclose

from ..spectra.spectrum import Spectrum, SpectralRegion
from ..manipulation.model_replace import model_replace
from ..fitting import fit_lines


def test_from_knots():
    wave_val = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    flux_val = np.array([2, 4, 6, 8, 10, 12, 14, 16, 18, 20])

    input_spectrum = Spectrum(spectral_axis=wave_val * u.AA, flux=flux_val * u.mJy)

    spline_knots = [3.5, 4.7, 6.8, 7.1] * u.AA

    # with default extrapolation, recovers the input flux
    result = model_replace(input_spectrum, None, model=spline_knots)

    assert result.uncertainty is None

    assert_quantity_allclose(result.flux, flux_val*u.mJy)
    assert_quantity_allclose(result.spectral_axis, input_spectrum.spectral_axis)

    # with zero fill extrapolation, fills with zeros.
    result = model_replace(input_spectrum, None, model=spline_knots, extrapolation_treatment='zero_fill')

    assert_quantity_allclose(result.flux[0], 0.*u.mJy)
    assert_quantity_allclose(result.flux[1], 0.*u.mJy)
    assert_quantity_allclose(result.flux[-1], 0.*u.mJy)
    assert_quantity_allclose(result.flux[-2], 0.*u.mJy)


def test_with_uncert_from_knots():
    wave_val = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    flux_val = np.array([2, 4, 6, 8, 10, 12, 14, 16, 18, 20])
    uncert_val = flux_val / 10.

    uncert = StdDevUncertainty(uncert_val * u.mJy)

    input_spectrum = Spectrum(spectral_axis=wave_val * u.AA, flux=flux_val * u.mJy,
                                uncertainty=uncert)

    spline_knots = [3.5, 4.7, 6.8, 7.1] * u.AA

    # When replacing directly at spline knots, the spectral region is
    # redundant and must be omitted.
    result = model_replace(input_spectrum, None, model=spline_knots)

    assert isinstance(result.uncertainty, StdDevUncertainty)
    assert result.flux.unit == result.uncertainty.unit

    assert_quantity_allclose(result.uncertainty.quantity, uncert_val*u.mJy)

    # Now try with the non-default no-uncertainty mode: result should
    # have no uncertainty even when input has.
    result = model_replace(input_spectrum, None, model=spline_knots,
                           interpolate_uncertainty=False)

    assert result.uncertainty is None


def test_with_uncert_zerofill_from_knots():
    wave_val = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    flux_val = np.array([2, 4, 6, 8, 10, 12, 14, 16, 18, 20])
    uncert_val = flux_val / 10.

    uncert = StdDevUncertainty(uncert_val * u.mJy)

    input_spectrum = Spectrum(spectral_axis=wave_val * u.AA, flux=flux_val * u.mJy,
                                uncertainty=uncert)

    spline_knots = [3.5, 4.7, 6.8, 7.1] * u.AA

    result = model_replace(input_spectrum, None, model=spline_knots,
                           extrapolation_treatment='zero_fill')

    assert isinstance(result.uncertainty, StdDevUncertainty)
    assert result.flux.unit == result.uncertainty.unit

    assert_quantity_allclose(result.uncertainty.quantity, uncert_val*u.mJy)


def test_from_region():
    wave_val = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    flux_val = np.array([2, 4, 6, 8, 10, 12, 14, 16, 18, 20])
    uncert_val = flux_val / 10.

    uncert = StdDevUncertainty(uncert_val * u.mJy)

    input_spectrum = Spectrum(spectral_axis=wave_val * u.AA, flux=flux_val * u.mJy,
                                uncertainty=uncert)

    region = SpectralRegion(3.5*u.AA, 7.1*u.AA)

    result = model_replace(input_spectrum, region, model=4)

    assert isinstance(result.uncertainty, StdDevUncertainty)
    assert result.flux.unit == result.uncertainty.unit
    assert_quantity_allclose(result.flux, flux_val*u.mJy)
    assert_quantity_allclose(result.spectral_axis, input_spectrum.spectral_axis)
    assert_quantity_allclose(result.uncertainty.quantity, uncert_val*u.mJy)


def test_from_fitted_model():
    wave_val = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    flux_val = np.array([1, 1.1, 0.9, 4., 10., 5., 2., 1., 1.2, 1.1])
    uncert_val = flux_val / 10.

    uncert = StdDevUncertainty(uncert_val * u.mJy)

    input_spectrum = Spectrum(spectral_axis=wave_val * u.AA, flux=flux_val * u.mJy,
                                uncertainty=uncert)

    model = models.Gaussian1D(10, 5.6, 1.2)
    fitted_model = fit_lines(input_spectrum, model)
    region = SpectralRegion(3.5*u.AA, 7.1*u.AA)

    result = model_replace(input_spectrum, region, model=fitted_model)

    assert result.uncertainty is None
    assert result.flux.unit == input_spectrum.flux.unit

    expected_flux = np.array([1, 1.1, 0.9, 4.408032, 9.582698,
                              5.612401, 0.885579, 1, 1.2, 1.1]) * u.mJy

    assert_quantity_allclose(result.flux, expected_flux)
    assert_quantity_allclose(result.spectral_axis, input_spectrum.spectral_axis)



import numpy as np
import pytest

import astropy.units as u
from astropy.modeling.polynomial import Chebyshev1D
from astropy.utils.exceptions import AstropyUserWarning

from ..spectra.spectrum import Spectrum
from ..spectra import SpectralRegion
from ..fitting.continuum import fit_generic_continuum, fit_continuum
from ..manipulation.smoothing import median_smooth


def single_peak_continuum(noise=0.2, constant_continuum=False):
    np.random.seed(0)
    x = np.linspace(0., 10., 200)
    y_single = 3 * np.exp(-0.5 * (x - 6.3)**2 / 0.1**2)
    y_single += np.random.normal(0., noise, x.shape)
    if not constant_continuum:
        y_continuum = 3.2 * np.exp(-0.5 * (x - 5.6)**2 / 4.8**2)
    else:
        y_continuum = 3.2
    y_single += y_continuum
    return x, y_single


def test_continuum_fit():
    """
    This test fits the first simulated spectrum from the fixture.  The
    initial guesses are manually set here with bounds that essentially make
    sense as the functionality of the test is to make sure the fit works and
    we get a reasonable answer out **given** good initial guesses.
    """

    x_single_continuum, y_single_continuum = single_peak_continuum()
    s_single_continuum = Spectrum(flux=y_single_continuum*u.Jy, spectral_axis=x_single_continuum*u.um)
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        g1_fit = fit_generic_continuum(s_single_continuum)

    y_continuum_fitted = g1_fit(s_single_continuum.spectral_axis)

    y_continuum_fitted_expected = np.array([1.71364056, 1.87755574, 2.05310622, 2.23545755, 2.41977527,
                                            2.60122493, 2.77497207, 2.93618225, 3.080021, 3.20165388,
                                            3.29624643, 3.3589642, 3.38497273, 3.36943758, 3.30752428,
                                            3.19439839, 3.02522545, 2.79517101, 2.49940062, 2.13307982])

    assert np.allclose(y_continuum_fitted.value[::10], y_continuum_fitted_expected, atol=1e-5)


def test_continuum_calculation():
    """
    This test fits the first simulated spectrum from the fixture.  The
    initial guesses are manually set here with bounds that essentially make
    sense as the functionality of the test is to make sure the fit works and
    we get a reasonable answer out **given** good initial guesses.
    """

    x_single_continuum, y_single_continuum = single_peak_continuum()
    spectrum = Spectrum(flux=y_single_continuum*u.Jy, spectral_axis=x_single_continuum*u.um)
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        g1_fit = fit_generic_continuum(spectrum)

    spectrum_normalized = spectrum / g1_fit(spectrum.spectral_axis)

    y_continuum_fitted_expected = np.array([1.15139925, 0.98509363, 0.73700614, 1.00911864, 0.913129,
                                            0.93145533, 0.94904202, 1.04162879, 0.90851397, 0.9494352,
                                            1.07812394, 1.06376489, 0.98705237, 0.94569623, 0.83502377,
                                            0.91909416, 0.89662208, 1.01458511, 0.96124191, 0.94847744])

    assert np.allclose(spectrum_normalized.flux.value[::10], y_continuum_fitted_expected, atol=1e-5)


def test_continuum_full_window():
    """
    This test fits the first simulated spectrum from the fixture, but
    with the fit_continuum function instead of fit_generic_continuum. Uses
    a window to select the entire spectrum and checks that it recovers the
    original, non-windowed fit.
    """

    x_single_continuum, y_single_continuum = single_peak_continuum()
    spectrum = Spectrum(flux=y_single_continuum*u.Jy, spectral_axis=x_single_continuum*u.um)

    # Smooth in the same way fit_generic_continuum does.
    spectrum_smoothed = median_smooth(spectrum, 3)

    # Check that a full width window recovers the original, non-windowed fit.
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        g1_fit = fit_continuum(spectrum_smoothed, window=(0.*u.um, 10.*u.um))
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        g1_fit_orig = fit_continuum(spectrum_smoothed)

    sp_normalized = spectrum / g1_fit(spectrum.spectral_axis)
    sp_normalized_orig = spectrum / g1_fit_orig(spectrum.spectral_axis)

    assert np.allclose(sp_normalized.flux.value, sp_normalized_orig.flux.value, atol=1e-5)


def test_continuum_spectral_region():
    """
    As before, but with a SpectralRegion as window.
    """

    x_single_continuum, y_single_continuum = single_peak_continuum()
    spectrum = Spectrum(flux=y_single_continuum*u.Jy, spectral_axis=x_single_continuum*u.um)

    # Smooth in the same way fit_generic_continuum does.
    spectrum_smoothed = median_smooth(spectrum, 3)

    # Check that a full width window recovers the original, non-windowed fit.
    region = SpectralRegion(0.*u.um, 10.*u.um)
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        g1_fit = fit_continuum(spectrum_smoothed, window=region)

    spectrum_normalized = spectrum / g1_fit(spectrum.spectral_axis)

    y_continuum_fitted_expected = np.array([1.15139925, 0.98509363, 0.73700614, 1.00911864, 0.913129,
                                            0.93145533, 0.94904202, 1.04162879, 0.90851397, 0.9494352,
                                            1.07812394, 1.06376489, 0.98705237, 0.94569623, 0.83502377,
                                            0.91909416, 0.89662208, 1.01458511, 0.96124191, 0.94847744])

    assert np.allclose(spectrum_normalized.flux.value[::10], y_continuum_fitted_expected, atol=1e-5)


def test_continuum_window_no_noise():
    """
    This test repeats the setup from the previous tests, but uses windows to
    select specific regions to fit. It uses a synthetic spectrum with no
    noise component. That way, numerical effects coming from the model fit
    itself can be more easily spotted.
    """

    x_single_continuum, y_single_continuum = single_peak_continuum(noise=0.)
    spectrum = Spectrum(flux=y_single_continuum*u.Jy, spectral_axis=x_single_continuum*u.um)

    # Smooth in the same way fit_generic_continuum does.
    spectrum_smoothed = median_smooth(spectrum, 3)

    # Window selects the first half of the spectrum.
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        g1_fit = fit_continuum(spectrum_smoothed, window=(0.*u.um, 5.*u.um))

    spectrum_normalized = spectrum / g1_fit(spectrum.spectral_axis)

    y_continuum_fitted_expected = np.ones(shape=(spectrum_normalized.spectral_axis.shape))

    # Check fit over the first half. The agreement is not so good as before,
    # probably due to the narrow component at 6.5 um.
    assert np.allclose(spectrum_normalized.flux.value[0:100], y_continuum_fitted_expected[0:100],
                       atol=5.5e-4)

    # Window selects the red end of the spectrum.
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        g1_fit = fit_continuum(spectrum_smoothed, window=(8.*u.um, 10.*u.um))

    spectrum_normalized = spectrum / g1_fit(spectrum.spectral_axis)

    # Check fit over the red end.
    assert np.allclose(spectrum_normalized.flux.value[160:], y_continuum_fitted_expected[160:],
                       atol=1.e-5)


def test_double_continuum_window():
    """
    Fit to no-noise spectrum comprised of a exponential continuum plus an emission Gaussian
    """
    x_single_continuum, y_single_continuum = single_peak_continuum(noise=0.,constant_continuum=False)
    spectrum = Spectrum(flux=y_single_continuum*u.Jy, spectral_axis=x_single_continuum*u.um)

    # Smooth in the same way fit_generic_continuum does.
    spectrum_smoothed = median_smooth(spectrum, 3)

    # Test spectrum is comprised of an exponential continuum that increases sharply at the
    # long wavelength end, plus a large-amplitude narrow Gaussian. We define the fitting
    # window as a SpectralRegion with sub-regions at the blue and red halves of the spectrum,
    # avoiding the Gaussian in between. The polynomial degree is high to accomodate the large
    # amplitude range of the expoinential continuum.
    region = SpectralRegion([(0. * u.um, 5. * u.um), (8. * u.um, 10. * u.um)])
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        g1_fit = fit_continuum(spectrum_smoothed, model=Chebyshev1D(7), window=region)

    spectrum_normalized = spectrum / g1_fit(spectrum.spectral_axis)

    y_continuum_fitted_expected = np.ones(shape=(spectrum_normalized.spectral_axis.shape))

    # Check fit over the windowed regions. Note that we fit a noiseless spectrum so we can
    # actually measure the degree of mismatch between model and data (polynomial X exponential).
    assert np.allclose(spectrum_normalized.flux.value[0:90], y_continuum_fitted_expected[0:90],
                       atol=1.e-4)
    assert np.allclose(spectrum_normalized.flux.value[170:], y_continuum_fitted_expected[170:],
                       atol=1.e-4)


def test_double_continuum_window_alternate():
    """
    Fit to no-noise spectrum comprised of a exponential continuum plus an emission Gaussian

    This is the same test as above, but with a different form for the double window
    specification.

    """
    x_single_continuum, y_single_continuum = single_peak_continuum(noise=0.,constant_continuum=False)
    spectrum = Spectrum(flux=y_single_continuum*u.Jy, spectral_axis=x_single_continuum*u.um)

    # Smooth in the same way fit_generic_continuum does.
    spectrum_smoothed = median_smooth(spectrum, 3)

    # Test spectrum is comprised of an exponential continuum that increases sharply at the
    # long wavelength end, plus a large-amplitude narrow Gaussian. We define the fitting
    # window as a SpectralRegion with sub-regions at the blue and red halves of the spectrum,
    # avoiding the Gaussian in between. The polynomial degree is high to accomodate the large
    # amplitude range of the expoinential continuum.
    region = [(0.*u.um, 5.*u.um), (8.* u.um, 10.* u.um)]
    with pytest.warns(AstropyUserWarning, match='Model is linear in parameters'):
        g1_fit = fit_continuum(spectrum_smoothed, model=Chebyshev1D(7), window=region)

    spectrum_normalized = spectrum / g1_fit(spectrum.spectral_axis)

    y_continuum_fitted_expected = np.ones(shape=(spectrum_normalized.spectral_axis.shape))

    # Check fit over the windowed regions. Note that we fit a noiseless spectrum so we can
    # actually measure the degree of mismatch between model and data (polynomial X exponential).
    assert np.allclose(spectrum_normalized.flux.value[0:90], y_continuum_fitted_expected[0:90],
                       atol=1.e-4)
    assert np.allclose(spectrum_normalized.flux.value[170:], y_continuum_fitted_expected[170:],
                       atol=1.e-4)



"""
A module for analysis tools focused on determining the location of
spectral features.
"""
import warnings

from astropy.nddata import StdDevUncertainty
from astropy.utils.exceptions import AstropyDeprecationWarning
import astropy.uncertainty as unc
import numpy as np

from ..spectra import SpectralRegion
from ..manipulation import extract_region


__all__ = ['centroid']


def centroid(spectrum, regions=None, region=None, analytic=True):
    """
    Calculate the centroid of a region, or regions, of the spectrum.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum.Spectrum`
        The spectrum object over which the centroid will be calculated.  If the uncertainty
        is populated, the returned quantity will include an uncertainty attribute with
        the propagated uncertainty (as Standard Deviation-style uncertainties).  This uncertainty
        assumes the input uncertainties are uncorrelated.

    regions : `~specutils.SpectralRegion` or list of `~specutils.SpectralRegion`
        Region within the spectrum to calculate the centroid.

    analytic : bool, optional
        Set this flag to ``False`` to use the `~astropy.uncertainty` distribution-based
        calculation for the centroid and its uncertainty instead of the default analytic solution.

    Returns
    -------
    centroid : float or list (based on region input)
        Centroid of the spectrum or within the regions

    Notes
    -----
    The spectrum will need to be continuum subtracted before calling
    this method. See the
    `analysis documentation <https://specutils.readthedocs.io/en/latest/analysis.html>`_ for more information.

    """

    if region is not None:
        regions = region
        warnings.warn("The 'region' keyword has been deprecated in favor "
                      "of 'regions' since specutils 1.8 and will be removed "
                      "in a future release.", AstropyDeprecationWarning)

    # No region, therefore whole spectrum.
    if regions is None:
        return _centroid_single_region(spectrum, analytic=analytic)

    # Single region
    elif isinstance(regions, SpectralRegion):
        return _centroid_single_region(spectrum, region=regions,
                                       analytic=analytic)

    # List of regions
    elif isinstance(regions, list):
        return [_centroid_single_region(spectrum, region=reg,
                                        analytic=analytic)
                for reg in regions]


def _centroid_single_region(spectrum, region=None, analytic=True):
    """
    Calculate the centroid of the spectrum based on the flux in the spectrum.
    The returned quantity object will have a ``.uncertainty`` attribute which
    will be populated if ``spectrum`` has uncertainties assigned, or ``None`` if not.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum.Spectrum`
        The spectrum object overwhich the centroid will be calculated.

    region : `~specutils.SpectralRegion`
        Region within the spectrum to calculate the centroid.

    analytic : bool, optional
        Set this flag to ``False`` to use the `~astropy.uncertainty` distribution-based
        calculation for the centroid and its uncertainty instead of the default analytic solution.

    Returns
    -------
    centroid : float or list (based on region input)
        Centroid of the spectrum or within the regions

    Notes
    -----
    This is a helper function for the above `centroid()` method.

    """
    if not analytic and spectrum.uncertainty is None:
        raise ValueError("Distribution-based calculation can only be used if"
                         " spectrum.uncertainty is not None")

    if region is not None:
        calc_spectrum = extract_region(spectrum, region)
    else:
        calc_spectrum = spectrum

    if spectrum.uncertainty is not None:
        flux_uncert = calc_spectrum.uncertainty.represent_as(StdDevUncertainty).quantity
    else:
        # dummy value for uncertainties to avoid extra if-statements when applying mask
        flux_uncert = np.zeros_like(calc_spectrum.flux)

    if hasattr(calc_spectrum, 'mask') and calc_spectrum.mask is not None:
        flux = calc_spectrum.flux[~calc_spectrum.mask]
        dispersion = calc_spectrum.spectral_axis[~calc_spectrum.mask].quantity
        flux_uncert = flux_uncert[~calc_spectrum.mask]
    else:
        flux = calc_spectrum.flux
        dispersion = calc_spectrum.spectral_axis.quantity

    if analytic:
        centroid = np.sum(flux*dispersion, axis=-1) / np.sum(flux, axis=-1)
        if spectrum.uncertainty is None:
            centroid.uncertainty = None
        else:
            N = np.sum(flux, axis=-1)
            # Looks overcomplicated, but gives us the right shape to match flux_uncert
            diff = np.subtract.outer(dispersion, centroid.transpose()).transpose()
            s2 = np.sum(flux_uncert*flux_uncert * diff*diff, axis=-1)/(N*N)

            # centroid uncertainty, fractionally added in quadrature of numerator and denom
            centroid.uncertainty = np.sqrt(s2).to(spectrum.spectral_axis.unit)
            centroid.uncertainty_type = 'stddev'
    else:
        # Convert flux to an astropy.uncertainties normal distribution
        flux = unc.normal(flux, std=flux_uncert, n_samples=1000)

        if len(flux.shape) > 1:
            dispersion = np.tile(dispersion, list(flux.shape[:-1]) + [1,])

        # centroid is the flux-weighted mean of the dispersions, the uncertainties
        # need to be scaled to the numerator/denominator, so we'll compute those in advance.

        # the axis=-2 will enable this to run on single-dispersion, single-flux
        # and single-dispersion, multiple-flux. Not -1 because of the distribution
        # n_samples axis
        numerator = (flux*dispersion).sum(axis=-2)
        denom = flux.sum(axis=-2)
        centroid_dist = numerator/denom
        centroid = centroid_dist.pdf_mean()
        if spectrum.uncertainty is None:
            centroid.uncertainty = None
        else:
            centroid.uncertainty = centroid_dist.pdf_std()
            centroid.uncertainty_type = 'stddev'

    return centroid



"""Contains classes that serialize spectral data types into ASDF representations."""
from asdf.extension import Converter
from asdf_astropy.converters import SpectralCoordConverter
from astropy.nddata import (StdDevUncertainty, VarianceUncertainty,
                            InverseVariance, UnknownUncertainty)

from specutils.spectra import Spectrum, SpectrumList

__all__ = ['SpectrumConverter', 'SpectrumListConverter']

UNCERTAINTY_TYPE_MAPPING = {
    'std': StdDevUncertainty,
    'var': VarianceUncertainty,
    'ivar': InverseVariance,
    'unknown': UnknownUncertainty}


class SpectralAxisConverter(SpectralCoordConverter):
    """ASDF converter to serialize/deserialize SpectralAxis objects."""
    tags = ["tag:astropy.org:specutils/spectra/spectral_axis-*"]
    types = ["specutils.spectra.spectral_axis.SpectralAxis"]

    def from_yaml_tree(self, node, tag, ctx):
        from specutils.spectra.spectral_axis import SpectralAxis

        return SpectralAxis(super().from_yaml_tree(node, tag, ctx))


class SpectrumConverter(Converter):
    """ASDF converter to serialize/deserialize Spectrum objects."""
    tags = ["tag:astropy.org:specutils/spectra/spectrum-*",
            "tag:astropy.org:specutils/spectra/spectrum1d-*"]
    types = ["specutils.spectra.spectrum.Spectrum",
             "specutils.spectra.spectrum.Spectrum1D"]

    def to_yaml_tree(self, obj, tag, ctx):
        """Converts Spectrum object into tree used for YAML representation."""
        node = {}
        node['flux'] = obj.flux
        node['spectral_axis'] = obj.spectral_axis

        if obj.uncertainty is not None:
            node['uncertainty'] = {}
            node['uncertainty']['uncertainty_type'] = obj.uncertainty.uncertainty_type
            data = obj.uncertainty.array
            node['uncertainty']['data'] = data

        if obj.mask is not None:
            node['mask'] = obj.mask

        return node

    def from_yaml_tree(cls, node, tag, ctx):
        """Converts tree representation back into Spectrum object."""
        flux = node['flux']
        spectral_axis = node['spectral_axis']
        uncertainty = node.get('uncertainty', None)
        mask = node.get('mask', None)

        if uncertainty is not None:
            class_ = UNCERTAINTY_TYPE_MAPPING[uncertainty['uncertainty_type']]
            data = uncertainty['data']
            uncertainty = class_(data)

        return Spectrum(flux=flux, spectral_axis=spectral_axis, uncertainty=uncertainty, mask=mask)


class SpectrumListConverter(Converter):
    """ASDF converter used to serialize/deserialize SpectrumList objects."""
    tags = ["tag:astropy.org:specutils/spectra/spectrum_list-*"]
    types = ["specutils.spectra.spectrum_list.SpectrumList"]

    def to_yaml_tree(self, obj, tag, ctx):
        """Converts SpectrumList object into tree used for YAML representation."""
        return [spectrum for spectrum in obj]

    def from_yaml_tree(cls, node, tag, ctx):
        """Converts tree representation back into SpectrumList object."""
        return SpectrumList(tree for tree in node)



"""
A module for analysis tools focused on determining fluxes of spectral features.
"""

import warnings
from functools import wraps

import numpy as np
from astropy.nddata import VarianceUncertainty

from .. import conf
from ..spectra import Spectrum
from ..manipulation import extract_region, LinearInterpolatedResampler
from .utils import computation_wrapper
import astropy.units as u
from astropy.stats import mad_std
from astropy.utils.exceptions import AstropyUserWarning


__all__ = ['line_flux', 'equivalent_width', 'is_continuum_below_threshold',
           'warn_continuum_below_threshold']


def line_flux(spectrum, regions=None,
              mask_interpolation=LinearInterpolatedResampler):
    """
    Computes the integrated flux in a spectrum or region of a spectrum.

    Applies to the whole spectrum by default, but can be limited to a specific
    feature (like a spectral line) if a region is given.

    Parameters
    ----------
    spectrum : Spectrum
        The spectrum object over which the summed flux will be calculated.

    regions : `~specutils.SpectralRegion` or list of `~specutils.SpectralRegion`
        Region within the spectrum to calculate the gaussian sigma width. If
        regions is `None`, computation is performed over entire spectrum.

    mask_interpolation : ``None`` or `~specutils.manipulation.LinearInterpolatedResampler`
        Interpolator class used to fill up the gaps in the spectrum's flux
        array, when the spectrum mask is not None. If set to ``None``, the
        masked spectral bins are excised from the data without interpolation
        and the bin edges of the adjacent bins are extended to fill the gap.

    Returns
    -------
    flux : `~astropy.units.Quantity`
        Flux in the provided spectrum (or regions). Unit isthe ``spectrum``'s'
        ``flux`` unit times ``spectral_axis`` unit.

    Notes
    -----
    While the flux can be computed on any spectrum or region, it should be
    continuum-subtracted to compute actual line fluxes.
    """
    return computation_wrapper(_compute_line_flux, spectrum, regions,
                               mask_interpolation=mask_interpolation)


def equivalent_width(spectrum, continuum=1, regions=None,
                     mask_interpolation=LinearInterpolatedResampler):
    """
    Computes the equivalent width of a region of the spectrum.

    Applies to the whole spectrum by default, but can be limited to a specific
    feature (like a spectral line) if a region is given.

    Parameters
    ----------
    spectrum : Spectrum
        The spectrum object overwhich the equivalent width will be calculated.

    regions: `~specutils.SpectralRegion` or list of `~specutils.SpectralRegion`
        Region within the spectrum to calculate the equivalent width. If
        regions is `None`, computation is performed over entire spectrum.

    continuum : ``1`` or `~astropy.units.Quantity`, optional
        Value to assume is the continuum level.  For the special value ``1``
        (without units), ``1`` in whatever the units of the ``spectrum.flux``
        will be assumed, otherwise units are required and must be the same as
        the ``spectrum.flux``.

    mask_interpolation : ``None`` or `~specutils.manipulation.LinearInterpolatedResampler`
        Interpolator class used to fill up the gaps in the spectrum's flux
        array after an excise operation to ensure the mask shape can always be
        applied when the spectrum mask is not None. If set to ``None``, the
        masked spectral bins are excised from the data without interpolation
        and the bin edges of the adjacent bins are extended to fill the gap.

    Returns
    -------
    ew : `~astropy.units.Quantity`
        Equivalent width calculation, in the same units as the ``spectrum``'s
        ``spectral_axis``.

    Notes
    -----
    To do a standard equivalent width measurement, the ``spectrum`` should be
    continuum-normalized to whatever ``continuum`` is before this function is
    called.

    """

    kwargs = dict(continuum=continuum)
    return computation_wrapper(_compute_equivalent_width, spectrum, regions,
                               mask_interpolation=mask_interpolation, **kwargs)


def _compute_line_flux(spectrum, regions=None,
                       mask_interpolation=LinearInterpolatedResampler):

    if regions is not None:
        calc_spectrum = extract_region(spectrum, regions)
    else:
        calc_spectrum = spectrum

    # Account for the existence of a mask.
    if hasattr(calc_spectrum, 'mask') and calc_spectrum.mask is not None:
        mask = calc_spectrum.mask
        new_spec = Spectrum(flux=calc_spectrum.flux[~mask],
                              spectral_axis=calc_spectrum.spectral_axis[~mask])

        if mask_interpolation is None:
            return _compute_line_flux(new_spec)
        else:
            interpolator = mask_interpolation(extrapolation_treatment='zero_fill')
            sp = interpolator(new_spec, calc_spectrum.spectral_axis)
            flux = sp.flux
    else:
        flux = calc_spectrum.flux

    dx = np.abs(np.diff(calc_spectrum.spectral_axis.bin_edges))
    line_flux = np.sum(flux * dx)

    line_flux.uncertainty = None
    line_flux.uncertainty_type = 'stddev'

    if calc_spectrum.uncertainty is not None:
        # Can't handle masks via interpolation here, since interpolators
        # only work with the flux array.
        try:
            variance_q = calc_spectrum.uncertainty.represent_as(VarianceUncertainty).quantity
        except ValueError:
            warnings.warn(f"Uncertainty type "
                          f"'{calc_spectrum.uncertainty.uncertainty_type}' "
                          f"was not recognized by line_flux. Proceeding "
                          f"without uncertainty in result.",
                          AstropyUserWarning)
            variance_q = None

        if variance_q is not None:
            line_flux.uncertainty = np.sqrt(
                np.sum(variance_q * dx**2))

    # TODO: we may want to consider converting to erg / cm^2 / sec by default
    return line_flux


def _compute_equivalent_width(spectrum, continuum=1, regions=None,
                              mask_interpolation=LinearInterpolatedResampler):
    if regions is not None:
        spectrum = extract_region(spectrum, regions)

    # Account for the existence of a mask.
    if hasattr(spectrum, 'mask') and spectrum.mask is not None:
        mask = spectrum.mask
        spectrum = Spectrum(
            flux=spectrum.flux[~mask],
            spectral_axis=spectrum.spectral_axis[~mask])

    # Calculate the continuum flux value
    continuum = u.Quantity(continuum, unit=spectrum.flux.unit)

    # If continuum is provided as a single scalar/quantity value, create an
    # array the same size as the processed spectrum. Otherwise, assume that the
    # continuum was provided as an array and use as-is.
    if continuum.size == 1:
        continuum = continuum * np.ones(spectrum.flux.size)

    cont_spec = Spectrum(flux=continuum,
                           spectral_axis=spectrum.spectral_axis)
    cont_flux = _compute_line_flux(cont_spec,
                                   mask_interpolation=mask_interpolation)
    line_flux = _compute_line_flux(spectrum,
                                   mask_interpolation=mask_interpolation)

    # Calculate equivalent width
    dx = np.abs(np.diff(spectrum.spectral_axis.bin_edges))
    ew = np.sum((1 - line_flux / cont_flux) * dx)

    # NOTE: uncertainty gets lost during unit translation, so we have to convert first
    ew = ew.to(spectrum.spectral_axis.unit)
    if line_flux.uncertainty is not None:
        # continuum has no uncertainty, here we assume no uncertainty on the bin edges (dx) as well
        # we need to add the line_flux uncertainties in quadrature for the sum above (which was
        # summed over the length of dx)
        ew.uncertainty = np.sqrt((line_flux.uncertainty.value**2)*len(dx)) * dx.unit
    else:
        ew.uncertainty = None

    ew.uncertainty_type = 'stddev'
    return ew


def is_continuum_below_threshold(spectrum, threshold=0.01):
    """
    Determine if the baseline of this spectrum is less than a threshold.
    I.e., an estimate of whether or not the continuum has been subtracted.

    If ``threshold`` is an `~astropy.units.Quantity` with flux units, this
    directly compares the median of the spectrum to the threshold.
    of the flux to the threshold.

    If the threshold is a float or dimensionless quantity then the spectrum's uncertainty will be
    used or an estimate of the uncertainty. If the uncertainty is present then the
    threshold is compared to the median of the flux divided by the
    uncertainty.  If the uncertainty is not present then the threshold
    is compared to the median of the flux divided by the
    `~astropy.stats.mad_std`.

    Parameters
    ----------
    spectrum : `~specutils.spectra.spectrum.Spectrum`
        The spectrum object over which the width will be calculated.

    threshold: float or `~astropy.units.Quantity`
        The tolerance on the quantification to confirm the continuum is
        near zero.

    Returns
    -------
    is_continuum_below_threshold: bool
        Return True if the continuum of the spectrum is below the threshold, False otherwise.

    """

    flux = spectrum.flux
    uncertainty = spectrum.uncertainty if hasattr(spectrum, 'uncertainty') else None

    # Apply the mask if it exists.
    if hasattr(spectrum, 'mask') and spectrum.mask is not None:
        flux = flux[~spectrum.mask]
        uncertainty = uncertainty[~spectrum.mask] if uncertainty else uncertainty

    # If the threshold has units then the assumption is that we want
    # to compare the median of the flux regardless of the
    # existence of the uncertainty.
    if hasattr(threshold, 'unit') and not threshold.unit == u.dimensionless_unscaled:
        return np.median(flux) < threshold

    # If threshold does not have a unit, ie it is not a quantity, then
    # we are going to calculate based on the S/N if the uncertainty
    # exists.
    if uncertainty and uncertainty.uncertainty_type != 'std':
        return np.median(flux / uncertainty.quantity) < threshold
    else:
        return np.median(flux) / mad_std(flux) < threshold


def warn_continuum_below_threshold(threshold=0.01):
    """
    Decorator for methods that should warn if the baseline
    of the spectrum does not appear to be below a threshold.

    The ``check`` parameter is based on the
    `astropy configuration system <http://docs.astropy.org/en/stable/config/#adding-new-configuration-items>`_.
    Examples are on that page to show how to turn off this type of warning checking.
    """
    def actual_decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            if conf.do_continuum_function_check:
                spectrum = args[0]
                if not is_continuum_below_threshold(spectrum, threshold):
                    if hasattr(threshold, 'unit'):
                        levelorsnr = 'value'
                    else:
                        levelorsnr = 'signal-to-noise'

                    message = "Spectrum is not below the threshold {} {}. ".format(levelorsnr, threshold)
                    message += "This may indicate you have not continuum subtracted this spectrum (or that you have but it has high SNR features).\n\n"
                    message += ("""If you want to suppress this warning either type """
                                """'specutils.conf.do_continuum_function_check = False' or """
                                """see http://docs.astropy.org/en/stable/config/#adding-new-configuration-items """
                                """for other ways to configure the warning.""")

                    warnings.warn(message, AstropyUserWarning)
            result = function(*args, **kwargs)
            return result
        return wrapper
    return actual_decorator



from astropy.nddata import VarianceUncertainty
from astropy.units import Unit
from astropy.wcs import WCS

from ...spectra import Spectrum, SpectrumList
from ..registers import data_loader
from ..parsing_utils import read_fileobj_or_hdulist


def identify_2dfgrs(origin, *args, **kwargs):
    """
    Identify if the current file is a 2dFGRS file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        return hdulist[0].header.get("IMAGE", "").strip() in ("SKYCHART", "SPECTRUM")


def load_spectrum_from_extension(hdu, primary_header):
    """
    2dFGRS files contain multiple spectra, this is a helper function to load the
    spectrum for a single HDU.
    """
    header = hdu.header
    # Due to the missing information, the WCS needs to be read in manually
    wcs = WCS(naxis=1)
    wcs.wcs.cdelt[0] = header["CDELT1"]
    wcs.wcs.crval[0] = header["CRVAL1"]
    wcs.wcs.crpix[0] = header["CRPIX1"]
    wcs.wcs.cunit[0] = Unit("Angstrom")

    meta = {
        "header": header,
        "primary_header": primary_header,
    }

    spectrum = hdu.data[0] * Unit("count/s")
    variance = VarianceUncertainty(hdu.data[1])
    sky = hdu.data[2] * Unit("count/s")

    meta["sky"] = Spectrum(flux=sky, wcs=wcs)

    return Spectrum(flux=spectrum, wcs=wcs, meta=meta, uncertainty=variance)


@data_loader(
    "2dFGRS", identifier=identify_2dfgrs, dtype=SpectrumList,
    extensions=["fit", "fits"], priority=10,
)
def twodfgrs_fits_loader(file_obj, **kwargs):
    """
    Load a file from the 2dF Galaxy Redshift Survey.

    The 2dF Galaxy Redshift Survey (2dFGRS) was a major spectroscopic survey
    taking full advantage of the unique capabilities of the 2dF facility built
    by the Anglo-Australian Observatory. The 2dFGRS is integrated with the 2dF
    QSO survey. Further details can be seen at http://www.2dfgrs.net/ or at
    https://docs.datacentral.org.au/.

    Parameters
    ----------
    file_obj: str or HDUList
        The path to the FITS file, or the loaded file as an HDUList
    Returns
    -------
    data: SpectrumList
        The 2dFGRS spectra represented by the data in this file.
    """

    spectra = []

    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        primary_header = hdulist[0].header

        for hdu in hdulist:
            if hdu.header.get("IMAGE", "").strip() == "SKYCHART":
                continue
            spectra.append(load_spectrum_from_extension(hdu, primary_header))

    return SpectrumList(spectra)



from astropy.units import Unit
from astropy.wcs import WCS

from ...spectra import Spectrum, SpectrumList
from ..registers import data_loader
from ..parsing_utils import read_fileobj_or_hdulist


def identify_2slaq_lrg(origin, *args, **kwargs):
    """
    Identify if the current file is a 2SLAQ-LRG file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        if hdulist[0].header["MSTITLE"].startswith("2dF-SDSS LRG/QSO survey"):
            # apparently the QSO part doesn't have MSTITLE, so we should be safe
            # with just the above condition, but just in case, we know they have
            # a different structure (LRG has one ext, QSO has more; LRG is 3d,
            # QSO is 1d
            if len(hdulist) == 1 and hdulist[0].data.ndim == 3:
                return True


@data_loader(
    "2SLAQ-LRG", identifier=identify_2slaq_lrg, dtype=SpectrumList,
    extensions=["fit", "fits"], priority=10,
)
def twoslaq_lrg_fits_loader(file_obj, **kwargs):
    """
    Load a file from the LRG subset of the 2dF-SDSS LRG/QSO survey (2SLAQ-LRG)
    file.

    2SLAQ was one of a number of surveys that used the 2dF instrument on the
    Anglo-Australian Telescope (AAT) at Siding Spring Observatory (SSO) near
    Coonabarabran, Australia, and followed up the 2QZ survey. Further details
    can be seen at http://www.physics.usyd.edu.au/2slaq/ or at
    https://docs.datacentral.org.au/.

    The LRG and QSO data appear to be in different formats, this only loads the
    LRG version. As there is a science and sky spectrum, the `SpectrumList`
    loader provides both, whereas the `Spectrum` loader only provides the
    science.

    Parameters
    ----------
    file_name: str
        The path to the FITS file
    Returns
    -------
    data: SpectrumList
        The 2SLAQ-LRG spectrum that is represented by the data in this file.
    """

    with read_fileobj_or_hdulist(file_obj, **kwargs) as hdulist:
        header = hdulist[0].header
        spectrum = hdulist[0].data[0, 0] * Unit("count/s")
        sky = hdulist[0].data[1, 0] * Unit("count/s")

    # Due to the odd structure of the file, the WCS needs to be read in
    # manually
    wcs = WCS(naxis=1)
    wcs.wcs.cdelt[0] = header["CD1_1"]
    wcs.wcs.crval[0] = header["CRVAL1"]
    wcs.wcs.crpix[0] = header["CRPIX1"]
    wcs.wcs.cunit[0] = Unit("Angstrom")

    meta = {"header": header}

    return SpectrumList([
        Spectrum(flux=spectrum, wcs=wcs, meta=meta),
        Spectrum(flux=sky, wcs=wcs, meta=meta),
    ])

# Commented out until discussion about whether to provide science-only or not
# @data_loader("2SLAQ-LRG", identifier=identify_2slaq_lrg,dtype=Spectrum,
#              extensions=["fit", "fits"])
# def twoslaq_lrg_fits_loader_only_science(filename, **kwargs):
#     """
#     Load a file from the LRG subset of the 2dF-SDSS LRG/QSO survey (2SLAQ-LRG)
#     file.
#
#     2SLAQ was one of a number of surveys that used the 2dF instrument on the
#     Anglo-Australian Telescope (AAT) at Siding Spring Observatory (SSO) near
#     Coonabarabran, Australia, and followed up the 2QZ survey. Further details
#     can be seen at http://www.physics.usyd.edu.au/2slaq/ or at
#     https://docs.datacentral.org.au/.
#
#     The LRG and QSO data appear to be in different formats, this only loads the
#     LRG version. As there is a science and sky spectrum, the `SpectrumList`
#     loader provides both, whereas the `Spectrum` loader only provides the
#     science.
#
#     Parameters
#     ----------
#     file_name: str
#         The path to the FITS file
#     Returns
#     -------
#     data: Spectrum
#         The 2SLAQ-LRG spectrum that is represented by the data in this file.
#     """
#     return SpectrumList.read(filename, format="2SLAQ-LRG")[0]



from specutils import SpectrumList
from specutils.io.registers import data_loader

from .dc_common import FITS_FILE_EXTS, SINGLE_SPLIT_LABEL
from ..parsing_utils import read_fileobj_or_hdulist

OZDES_CONFIG = {
    "hdus": {
        "0": {"purpose": "combined_science"},
        "1": {"purpose": "combined_error_variance"},
        "2": {"purpose": "skip"},
        "cycle": {
            "0": {"purpose": "science"},
            "1": {"purpose": "error_variance"},
            "2": {"purpose": "skip"},
        },
    },
    "units": None,
    "wcs": None,
    "all_standard_units": True,
    "all_keywords": False,
    "valid_wcs": True,
}


def identify_ozdes(origin, *args, **kwargs):
    """
    Identify if the current file is a OzDES file
    """
    with read_fileobj_or_hdulist(*args, **kwargs) as hdulist:
        if "ozdes" in hdulist[0].header.get("REFERENC"):
            return True
        return False


@data_loader(
    label="OzDES", extensions=FITS_FILE_EXTS, dtype=SpectrumList,
    identifier=identify_ozdes, priority=10
)
def ozdes_loader(filename):
    spectra = SpectrumList.read(
        filename, format=SINGLE_SPLIT_LABEL, **OZDES_CONFIG
    )
    return spectra



from astropy.nddata import NDIOMixin


__all__ = ['SpectrumList']


class SpectrumList(list, NDIOMixin):
    """
    A list that is used to hold a list of `~specutils.Spectrum` objects

    The primary purpose of this class is to allow loaders to return a list of
    spectra that have different shapes.  For spectra that have the same shape
    but different spectral axes, see `~specutils.SpectrumCollection`.  For
    a spectrum or spectra that all share the same spectral axis, use
    `~specutils.Spectrum`.  For more on this topic, see
    :ref:`specutils-representation-overview`.
    """



"""
A module for analysis tools dealing with uncertainties or error analysis in
spectra.
"""

import copy
import operator

from astropy.nddata import StdDevUncertainty

__all__ = ['snr_threshold']


def snr_threshold(spectrum, value, op=operator.gt):
    """
    Calculate the mean S/N of the spectrum based on the flux and uncertainty
    in the spectrum. This will be calculated over the regions, if they
    are specified.

    Parameters
    ----------
    spectrum : `~specutils.Spectrum`, `~specutils.SpectrumCollection` or `~astropy.nddata.NDData`
        The spectrum object overwhich the S/N threshold will be calculated.

    value: ``float``
        Threshold value to be applied to flux / uncertainty.

    op: One of operator.gt, operator.ge, operator.lt, operator.le or
        the str equivalent '>', '>=', '<', '<='
        The mathematical operator to apply for thresholding.

    Returns
    -------
    spectrum: `~specutils.Spectrum`
        Output object with ``spectrum.mask`` set based on threshold.

    Notes
    -----
    The input object will need to have the uncertainty defined in order for the SNR
    to be calculated.

    """

    # Setup the mapping
    operator_mapping = {
        '>': operator.gt,
        '<': operator.lt,
        '>=': operator.ge,
        '<=': operator.le
    }

    if not hasattr(spectrum, 'uncertainty') or spectrum.uncertainty is None:
        raise Exception("S/N thresholding requires the uncertainty be defined.")

    if (op not in [operator.gt, operator.ge, operator.lt, operator.le] and
            op not in operator_mapping.keys()):
        raise ValueError('Threshold operator must be a string or operator that represents ' +
                         'greater-than, less-than, greater-than-or-equal or ' +
                         'less-than-or-equal')

    # If the operator passed in is a string, then map to the
    # operator method.
    if isinstance(op, str):
        op = operator_mapping[op]

    # Spectrum
    if hasattr(spectrum, 'flux'):
        data = spectrum.flux

    # NDData
    elif hasattr(spectrum, 'data'):
        data = spectrum.data * (spectrum.unit if spectrum.unit is not None else 1)
    else:
        raise ValueError('Could not find data attribute.')

    # NDData convention: Masks should follow the numpy convention that valid
    # data points are marked by False and invalid ones with True.
    mask = ~op(data / (spectrum.uncertainty.represent_as(StdDevUncertainty).quantity), value)

    spectrum_out = copy.copy(spectrum)
    spectrum_out._mask = mask

    return spectrum_out
